# Copyright 2014
# The Cloudscaling Group, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import netutils

from ec2api.api import common
from ec2api.api import ec2utils
from ec2api import clients
from ec2api import exception


availability_zone_opts = [
    cfg.StrOpt('internal_service_availability_zone',
               default='internal',
               help='The availability_zone to show internal services under'),
    cfg.StrOpt('my_ip',
               default=netutils.get_my_ipv4(),
               help='IP address of this host'),
    cfg.StrOpt('ec2_host',
               default='$my_ip',
               help='The IP address of the EC2 API server'),
    cfg.IntOpt('ec2_port',
               default=8788,
               help='The port of the EC2 API server'),
    cfg.StrOpt('ec2_scheme',
               default='http',
               help='The protocol to use when connecting to the EC2 API '
                    'server (http, https)'),
    cfg.StrOpt('ec2_path',
               default='/',
               help='The path prefix used to call the ec2 API server'),
    cfg.ListOpt('region_list',
                default=[],
                help='List of region=fqdn pairs separated by commas'),
]

CONF = cfg.CONF
CONF.register_opts(availability_zone_opts)
LOG = logging.getLogger(__name__)

"""Availability zones, regions, account attributes related API implementation
"""


Validator = common.Validator


def get_account_attribute_engine():
    return AccountAttributeEngineNeutron()


class AvailabilityZoneDescriber(common.UniversalDescriber):

    KIND = 'az'
    FILTER_MAP = {'state': 'zoneState',
                  'zone-name': 'zoneName'}

    def format(self, item=None, os_item=None):
        return _format_availability_zone(os_item)

    def get_db_items(self):
        return []

    def get_os_items(self):
        nova = clients.nova(self.context)
        zones = nova.availability_zones.list(detailed=False)
        for zone in zones:
            if zone.zoneName == CONF.internal_service_availability_zone:
                zones.remove(zone)
        return zones

    def get_name(self, os_item):
        return os_item.zoneName

    def get_id(self, os_item):
        return ''

    def auto_update_db(self, item, os_item):
        pass


def describe_availability_zones(context, zone_name=None, filter=None):
    # NOTE(Alex): Openstack extension, AWS-incompability
    # Checking for 'verbose' in zone_name.
    if zone_name and 'verbose' in zone_name:
        return _describe_verbose(context)

    formatted_availability_zones = AvailabilityZoneDescriber().describe(
        context, names=zone_name, filter=filter)
    return {'availabilityZoneInfo': formatted_availability_zones}


def describe_regions(context, region_name=None, filter=None):
    # TODO(andrey-mp): collect regions from keystone catalog
    if CONF.region_list:
        regions = []
        for region in CONF.region_list:
            name, _sep, host = region.partition('=')
            if not host:
                host = CONF.ec2_host
            endpoint = '%s://%s:%s%s' % (CONF.ec2_scheme,
                                         host,
                                         CONF.ec2_port,
                                         CONF.ec2_path)
            regions.append({'regionName': name,
                            'regionEndpoint': endpoint})
    else:
        # NOTE(andrey-mp): RegionOne is a default region name that is used
        # in keystone, nova and some other projects
        regions = [{'regionName': 'RegionOne',
                    'regionEndpoint': '%s://%s:%s%s' % (CONF.ec2_scheme,
                                                        CONF.ec2_host,
                                                        CONF.ec2_port,
                                                        CONF.ec2_path)}]
    return {'regionInfo': regions}


def describe_account_attributes(context, attribute_name=None):
    def get_max_instances():
        nova = clients.nova(context)
        quotas = nova.quotas.get(context.project_id, context.user_id)
        return quotas.instances

    attribute_getters = {
        'supported-platforms': (
            account_attribute_engine.get_supported_platforms),
        'default-vpc': functools.partial(
            account_attribute_engine.get_default_vpc, context),
        'max-instances': get_max_instances,
    }

    formatted_attributes = []
    for attribute in (attribute_name or attribute_getters):
        if attribute not in attribute_getters:
            raise exception.InvalidParameter(name=attribute)
        formatted_attributes.append(
            _format_account_attribute(attribute,
                                      attribute_getters[attribute]()))
    return {'accountAttributeSet': formatted_attributes}


def _format_availability_zone(zone):
    return {'zoneName': zone.zoneName,
            'zoneState': ('available'
                          if zone.zoneState.get('available')
                          else 'unavailable')
            }


def _format_account_attribute(attribute, value):
    if not isinstance(value, list):
        value = [value]
    return {'attributeName': attribute,
            'attributeValueSet': [{'attributeValue': val} for val in value]}


# NOTE(Alex): Openstack extension, AWS-incompability
# The whole function and its result is incompatible with AWS.

def _describe_verbose(context):
    nova = clients.nova(context)
    availability_zones = nova.availability_zones.list()

    formatted_availability_zones = []
    for availability_zone in availability_zones:
        formatted_availability_zones.append(
            _format_availability_zone(availability_zone))
        for host, services in availability_zone.hosts.items():
            formatted_availability_zones.append(
                {'zoneName': '|- %s' % host,
                 'zoneState': ''})
            for service, values in services.items():
                active = ":-)" if values['active'] else "XXX"
                enabled = 'enabled' if values['available'] else 'disabled'
                formatted_availability_zones.append(
                    {'zoneName': '| |- %s' % service,
                     'zoneState': ('%s %s %s' % (enabled, active,
                                                 values['updated_at']))})

    return {'availabilityZoneInfo': formatted_availability_zones}


class AccountAttributeEngineNeutron(object):

    def get_supported_platforms(self):
        if CONF.disable_ec2_classic:
            return ['VPC']
        else:
            return ['EC2', 'VPC']

    def get_default_vpc(self, context):
        if CONF.disable_ec2_classic:
            default_vpc = ec2utils.check_and_create_default_vpc(context)
            if default_vpc:
                return default_vpc['id']
        return 'none'


account_attribute_engine = get_account_attribute_engine()



# Copyright 2014
# The Cloudscaling Group, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64

from novaclient import exceptions as nova_exception
from oslo_config import cfg
from oslo_log import log as logging

from ec2api.api import common
from ec2api import clients
from ec2api import exception
from ec2api.i18n import _


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


"""Keypair-object related API implementation
"""


Validator = common.Validator


class KeyPairDescriber(common.UniversalDescriber):

    KIND = 'kp'
    FILTER_MAP = {'fingerprint': 'keyFingerprint',
                  'key-name': 'keyName'}

    def format(self, _item, key_pair):
        return _format_key_pair(key_pair)

    def get_db_items(self):
        return []

    def get_os_items(self):
        # Original EC2 in nova filters out vpn keys for admin user.
        # We're not filtering out the vpn keys for now.
        # In order to implement this we'd have to configure vpn_key_suffix
        # in our config which we consider an overkill.
        # suffix = CONF.vpn_key_suffix
        # if context.is_admin or not key_pair['name'].endswith(suffix):
        nova = clients.nova(self.context)
        return nova.keypairs.list()

    def auto_update_db(self, item, os_item):
        pass

    def get_id(self, os_item):
        return ''

    def get_name(self, key_pair):
        return key_pair.name


def describe_key_pairs(context, key_name=None, filter=None):
    formatted_key_pairs = KeyPairDescriber().describe(context, names=key_name,
                                                      filter=filter)
    return {'keySet': formatted_key_pairs}


def _validate_name(name):
    if len(name) > 255:
        raise exception.InvalidParameterValue(
            value=name,
            parameter='KeyName',
            reason='lenght is exceeds maximum of 255')


def create_key_pair(context, key_name):
    _validate_name(key_name)
    nova = clients.nova(context)
    try:
        key_pair = nova.keypairs.create(key_name)
    except nova_exception.OverLimit:
        raise exception.ResourceLimitExceeded(resource='keypairs')
    except nova_exception.Conflict:
        raise exception.InvalidKeyPairDuplicate(key_name=key_name)
    formatted_key_pair = _format_key_pair(key_pair)
    formatted_key_pair['keyMaterial'] = key_pair.private_key
    return formatted_key_pair


def import_key_pair(context, key_name, public_key_material):
    _validate_name(key_name)
    if not public_key_material:
        raise exception.MissingParameter(
            _('The request must contain the parameter PublicKeyMaterial'))
    nova = clients.nova(context)
    public_key = base64.b64decode(public_key_material).decode("utf-8")
    try:
        key_pair = nova.keypairs.create(key_name, public_key)
    except nova_exception.OverLimit:
        raise exception.ResourceLimitExceeded(resource='keypairs')
    except nova_exception.Conflict:
        raise exception.InvalidKeyPairDuplicate(key_name=key_name)
    return _format_key_pair(key_pair)


def delete_key_pair(context, key_name):
    nova = clients.nova(context)
    try:
        nova.keypairs.delete(key_name)
    except nova_exception.NotFound:
        # aws returns true even if the key doesn't exist
        pass
    return True


def _format_key_pair(key_pair):
    return {'keyName': key_pair.name,
            'keyFingerprint': key_pair.fingerprint
            }



# Copyright 2014
# The Cloudscaling Group, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from unittest import mock

from neutronclient.common import exceptions as neutron_exception

from ec2api.api import common
from ec2api.api import vpn_connection as vpn_connection_api
from ec2api.tests.unit import base
from ec2api.tests.unit import fakes
from ec2api.tests.unit import matchers
from ec2api.tests.unit import tools


class VpnConnectionTestCase(base.ApiTestCase):

    @mock.patch('ec2api.api.vpn_connection.describe_vpn_connections')
    @mock.patch('ec2api.api.vpn_connection._reset_vpn_connections',
                wraps=vpn_connection_api._reset_vpn_connections)
    @mock.patch('random.choice')
    def test_create_vpn_connection(self, random_choice, reset_vpn_connections,
                                   describe_vpn_connections):
        self.set_mock_db_items(
            fakes.DB_VPN_GATEWAY_1, fakes.DB_VPN_GATEWAY_2,
            fakes.DB_CUSTOMER_GATEWAY_1, fakes.DB_CUSTOMER_GATEWAY_2,
            fakes.DB_VPC_1)
        self.neutron.create_ikepolicy.side_effect = (
            tools.get_neutron_create('ikepolicy', fakes.ID_OS_IKEPOLICY_1))
        self.neutron.create_ipsecpolicy.side_effect = (
            tools.get_neutron_create('ipsecpolicy', fakes.ID_OS_IPSECPOLICY_1))
        self.db_api.add_item.side_effect = (
            tools.get_db_api_add_item(fakes.ID_EC2_VPN_CONNECTION_1))
        random_choice.side_effect = iter(fakes.PRE_SHARED_KEY_1)
        describe_vpn_connections.return_value = {
            'vpnConnectionSet': [fakes.EC2_VPN_CONNECTION_1]}

        resp = self.execute(
            'CreateVpnConnection',
            {'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_1,
             'CustomerGatewayId': fakes.ID_EC2_CUSTOMER_GATEWAY_1,
             'Type': 'ipsec.1',
             'Options.StaticRoutesOnly': 'True'})
        self.assertThat(
            resp,
            matchers.DictMatches(
                {'vpnConnection': fakes.EC2_VPN_CONNECTION_1}))

        self.neutron.create_ikepolicy.assert_called_once_with(
            {'ikepolicy': tools.purge_dict(fakes.OS_IKEPOLICY_1, ('id',))})
        self.neutron.create_ipsecpolicy.assert_called_once_with(
            {'ipsecpolicy': tools.purge_dict(fakes.OS_IPSECPOLICY_1, ('id',))})
        random_choice.assert_called_with(vpn_connection_api.SHARED_KEY_CHARS)
        new_vpn_connection_1 = tools.update_dict(
            fakes.DB_VPN_CONNECTION_1, {'cidrs': [],
                                        'os_ipsec_site_connections': {}})
        self.db_api.add_item.assert_called_once_with(
            mock.ANY, 'vpn',
            tools.purge_dict(new_vpn_connection_1, ('id', 'vpc_id', 'os_id')))
        self.neutron.update_ikepolicy.assert_called_once_with(
            fakes.ID_OS_IKEPOLICY_1,
            {'ikepolicy': {'name': fakes.ID_EC2_VPN_CONNECTION_1}})
        self.neutron.update_ipsecpolicy.assert_called_once_with(
            fakes.ID_OS_IPSECPOLICY_1,
            {'ipsecpolicy': {'name': fakes.ID_EC2_VPN_CONNECTION_1}})
        reset_vpn_connections.assert_called_once_with(
            mock.ANY, self.neutron, mock.ANY, fakes.DB_VPN_GATEWAY_1,
            vpn_connections=[new_vpn_connection_1])
        self.assertIsInstance(reset_vpn_connections.call_args[0][2],
                              common.OnCrashCleaner)
        describe_vpn_connections.assert_called_once_with(
            mock.ANY, vpn_connection_id=[fakes.ID_EC2_VPN_CONNECTION_1])

    @mock.patch('ec2api.api.vpn_connection.describe_vpn_connections')
    def test_create_vpn_connection_idempotent(self, describe_vpn_connections):
        self.set_mock_db_items(
            fakes.DB_VPN_GATEWAY_1, fakes.DB_CUSTOMER_GATEWAY_1,
            fakes.DB_VPN_CONNECTION_1)
        describe_vpn_connections.return_value = {
            'vpnConnectionSet': [fakes.EC2_VPN_CONNECTION_1]}

        resp = self.execute(
            'CreateVpnConnection',
            {'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_1,
             'CustomerGatewayId': fakes.ID_EC2_CUSTOMER_GATEWAY_1,
             'Type': 'ipsec.1',
             'Options.StaticRoutesOnly': 'True'})
        self.assertThat({'vpnConnection': fakes.EC2_VPN_CONNECTION_1},
                        matchers.DictMatches(resp))
        self.assertFalse(self.neutron.create_ikepolicy.called)
        self.assertFalse(self.neutron.create_ipsecpolicy.called)
        self.assertFalse(self.db_api.add_item.called)
        describe_vpn_connections.assert_called_once_with(
            mock.ANY, vpn_connection_id=[fakes.ID_EC2_VPN_CONNECTION_1])

    def test_create_vpn_connection_invalid_parameters(self):
        self.assert_execution_error(
            'Unsupported', 'CreateVpnConnection',
            {'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_1,
             'CustomerGatewayId': fakes.ID_EC2_CUSTOMER_GATEWAY_1,
             'Type': 'ipsec.1',
             'Options.StaticRoutesOnly': 'False'})

        self.assert_execution_error(
            'Unsupported', 'CreateVpnConnection',
            {'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_1,
             'CustomerGatewayId': fakes.ID_EC2_CUSTOMER_GATEWAY_1,
             'Type': 'ipsec.1'})

        self.set_mock_db_items(fakes.DB_CUSTOMER_GATEWAY_1)
        self.assert_execution_error(
            'InvalidVpnGatewayID.NotFound', 'CreateVpnConnection',
            {'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_2,
             'CustomerGatewayId': fakes.ID_EC2_CUSTOMER_GATEWAY_1,
             'Type': 'ipsec.1',
             'Options.StaticRoutesOnly': 'True'})

        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_1)
        self.assert_execution_error(
            'InvalidCustomerGatewayID.NotFound', 'CreateVpnConnection',
            {'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_2,
             'CustomerGatewayId': fakes.ID_EC2_CUSTOMER_GATEWAY_1,
             'Type': 'ipsec.1',
             'Options.StaticRoutesOnly': 'True'})

        self.set_mock_db_items(
            fakes.DB_VPN_GATEWAY_2, fakes.DB_CUSTOMER_GATEWAY_1,
            fakes.DB_VPN_CONNECTION_1)
        self.assert_execution_error(
            'InvalidCustomerGateway.DuplicateIpAddress', 'CreateVpnConnection',
            {'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_2,
             'CustomerGatewayId': fakes.ID_EC2_CUSTOMER_GATEWAY_1,
             'Type': 'ipsec.1',
             'Options.StaticRoutesOnly': 'True'})

    @tools.screen_unexpected_exception_logs
    def test_create_vpn_connection_rollback(self):
        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_1,
                               fakes.DB_CUSTOMER_GATEWAY_1)
        self.neutron.create_ikepolicy.side_effect = (
            tools.get_neutron_create('ikepolicy', fakes.ID_OS_IKEPOLICY_1))
        self.neutron.create_ipsecpolicy.side_effect = (
            tools.get_neutron_create('ipsecpolicy', fakes.ID_OS_IPSECPOLICY_1))
        self.db_api.add_item.side_effect = (
            tools.get_db_api_add_item(fakes.ID_EC2_VPN_CONNECTION_1))
        self.neutron.update_ikepolicy.side_effect = Exception()

        self.assert_execution_error(
            self.ANY_EXECUTE_ERROR, 'CreateVpnConnection',
            {'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_1,
             'CustomerGatewayId': fakes.ID_EC2_CUSTOMER_GATEWAY_1,
             'Type': 'ipsec.1',
             'Options.StaticRoutesOnly': 'True'})

        self.db_api.delete_item.assert_called_once_with(
            mock.ANY, fakes.ID_EC2_VPN_CONNECTION_1)
        self.neutron.delete_ipsecpolicy.assert_called_once_with(
            fakes.ID_OS_IPSECPOLICY_1)
        self.neutron.delete_ikepolicy.assert_called_once_with(
            fakes.ID_OS_IKEPOLICY_1)

    @mock.patch('ec2api.api.vpn_connection._reset_vpn_connections',
                wraps=vpn_connection_api._reset_vpn_connections)
    def test_create_vpn_connection_route(self, reset_vpn_connections):
        self.set_mock_db_items(fakes.DB_VPN_CONNECTION_2,
                               fakes.DB_VPN_GATEWAY_2)

        resp = self.execute(
            'CreateVpnConnectionRoute',
            {'VpnConnectionId': fakes.ID_EC2_VPN_CONNECTION_2,
             'DestinationCidrBlock': '192.168.123.0/24'})
        self.assertEqual({'return': True}, resp)

        vpn = copy.deepcopy(fakes.DB_VPN_CONNECTION_2)
        vpn['cidrs'].append('192.168.123.0/24')
        self.db_api.update_item.assert_called_once_with(mock.ANY, vpn)
        reset_vpn_connections.assert_called_once_with(
            mock.ANY, self.neutron, mock.ANY, fakes.DB_VPN_GATEWAY_2,
            vpn_connections=[vpn])

    def test_create_vpn_connection_route_idempotent(self):
        self.set_mock_db_items(fakes.DB_VPN_CONNECTION_2)

        resp = self.execute(
            'CreateVpnConnectionRoute',
            {'VpnConnectionId': fakes.ID_EC2_VPN_CONNECTION_2,
             'DestinationCidrBlock': fakes.CIDR_VPN_2_PROPAGATED_1})
        self.assertEqual({'return': True}, resp)
        self.assertFalse(self.db_api.update_item.called)

    def test_create_vpn_connection_route_invalid_parameters(self):
        self.set_mock_db_items()
        self.assert_execution_error(
            'InvalidVpnConnectionID.NotFound', 'CreateVpnConnectionRoute',
            {'VpnConnectionId': fakes.ID_EC2_VPN_CONNECTION_2,
             'DestinationCidrBlock': fakes.CIDR_VPN_2_PROPAGATED_1})

    @tools.screen_unexpected_exception_logs
    @mock.patch('ec2api.api.vpn_connection._reset_vpn_connections')
    def test_create_vpn_connection_route_rollback(self, reset_vpn_connections):
        self.set_mock_db_items(fakes.DB_VPN_CONNECTION_2,
                               fakes.DB_VPN_GATEWAY_2)
        reset_vpn_connections.side_effect = Exception()

        self.assert_execution_error(
            self.ANY_EXECUTE_ERROR, 'CreateVpnConnectionRoute',
            {'VpnConnectionId': fakes.ID_EC2_VPN_CONNECTION_2,
             'DestinationCidrBlock': '192.168.123.0/24'})
        self.db_api.update_item.assert_called_with(
            mock.ANY, fakes.DB_VPN_CONNECTION_2)

    @mock.patch('ec2api.api.vpn_connection._reset_vpn_connections',
                wraps=vpn_connection_api._reset_vpn_connections)
    def test_delete_vpn_connection_route(self, reset_vpn_connections):
        self.set_mock_db_items(fakes.DB_VPN_CONNECTION_2,
                               fakes.DB_VPN_GATEWAY_2)

        resp = self.execute(
            'DeleteVpnConnectionRoute',
            {'VpnConnectionId': fakes.ID_EC2_VPN_CONNECTION_2,
             'DestinationCidrBlock': fakes.CIDR_VPN_2_PROPAGATED_1})
        self.assertEqual({'return': True}, resp)
        vpn = tools.update_dict(fakes.DB_VPN_CONNECTION_2,
                                {'cidrs': [fakes.CIDR_VPN_2_PROPAGATED_2]})
        self.db_api.update_item.assert_called_once_with(mock.ANY, vpn)
        reset_vpn_connections.assert_called_once_with(
            mock.ANY, self.neutron, mock.ANY, fakes.DB_VPN_GATEWAY_2,
            vpn_connections=[vpn])

    def test_delete_vpn_connection_route_invalid_parameters(self):
        self.set_mock_db_items()
        self.assert_execution_error(
            'InvalidVpnConnectionID.NotFound', 'DeleteVpnConnectionRoute',
            {'VpnConnectionId': fakes.ID_EC2_VPN_CONNECTION_2,
             'DestinationCidrBlock': fakes.CIDR_VPN_2_PROPAGATED_1})

        self.set_mock_db_items(fakes.DB_VPN_CONNECTION_2)
        self.assert_execution_error(
            'InvalidRoute.NotFound', 'DeleteVpnConnectionRoute',
            {'VpnConnectionId': fakes.ID_EC2_VPN_CONNECTION_2,
             'DestinationCidrBlock': '192.168.123.0/24'})

    @tools.screen_unexpected_exception_logs
    @mock.patch('ec2api.api.vpn_connection._reset_vpn_connections')
    def test_delete_vpn_connection_route_rollback(self, reset_vpn_connections):
        self.set_mock_db_items(fakes.DB_VPN_CONNECTION_2,
                               fakes.DB_VPN_GATEWAY_2)
        reset_vpn_connections.side_effect = Exception()

        self.assert_execution_error(
            self.ANY_EXECUTE_ERROR, 'DeleteVpnConnectionRoute',
            {'VpnConnectionId': fakes.ID_EC2_VPN_CONNECTION_2,
             'DestinationCidrBlock': fakes.CIDR_VPN_2_PROPAGATED_1})
        self.assert_any_call(self.db_api.update_item,
                             mock.ANY, fakes.DB_VPN_CONNECTION_2)

    def test_delete_vpn_connection(self):
        self.set_mock_db_items(fakes.DB_VPN_CONNECTION_1)
        resp = self.execute('DeleteVpnConnection',
                            {'VpnConnectionId': fakes.ID_EC2_VPN_CONNECTION_1})
        self.assertEqual({'return': True}, resp)
        self.db_api.delete_item.assert_called_once_with(
            mock.ANY, fakes.ID_EC2_VPN_CONNECTION_1)
        self.neutron.delete_ipsec_site_connection.assert_called_once_with(
            fakes.ID_OS_IPSEC_SITE_CONNECTION_2)
        self.neutron.delete_ipsecpolicy.assert_called_once_with(
            fakes.ID_OS_IPSECPOLICY_1)
        self.neutron.delete_ikepolicy.assert_called_once_with(
            fakes.ID_OS_IKEPOLICY_1)

    def test_delete_vpn_connection_invalid_parameters(self):
        self.set_mock_db_items()
        self.assert_execution_error(
            'InvalidVpnConnectionID.NotFound', 'DeleteVpnConnection',
            {'VpnConnectionId': fakes.ID_EC2_VPN_CONNECTION_1})

    @tools.screen_unexpected_exception_logs
    def test_delete_vpn_connection_rollback(self):
        self.set_mock_db_items(fakes.DB_VPN_CONNECTION_1)
        self.neutron.delete_ikepolicy.side_effect = Exception()

        self.assert_execution_error(
            self.ANY_EXECUTE_ERROR, 'DeleteVpnConnection',
            {'VpnConnectionId': fakes.ID_EC2_VPN_CONNECTION_1})

        self.db_api.restore_item.assert_called_once_with(
            mock.ANY, 'vpn', fakes.DB_VPN_CONNECTION_1)
        self.assertFalse(self.neutron.create_ipsec_site_connection.called)
        self.assertFalse(self.neutron.create_ipsecpolicy.called)
        self.assertFalse(self.neutron.create_ikepolicy.called)

    def test_describe_vpn_connections(self):
        self.set_mock_db_items(
            fakes.DB_VPN_CONNECTION_1, fakes.DB_VPN_CONNECTION_2,
            fakes.DB_CUSTOMER_GATEWAY_1, fakes.DB_CUSTOMER_GATEWAY_2,
            fakes.DB_VPN_GATEWAY_1, fakes.DB_VPN_GATEWAY_2,
            fakes.DB_VPC_1, fakes.DB_VPC_2)
        self.neutron.list_ikepolicies.return_value = {
            'ikepolicies': [fakes.OS_IKEPOLICY_1, fakes.OS_IKEPOLICY_2]}
        self.neutron.list_ipsecpolicies.return_value = {
            'ipsecpolicies': [fakes.OS_IPSECPOLICY_1, fakes.OS_IPSECPOLICY_2]}
        self.neutron.list_ipsec_site_connections.return_value = {
            'ipsec_site_connections': []}
        self.neutron.list_routers.return_value = {
            'routers': [fakes.OS_ROUTER_1, fakes.OS_ROUTER_2]}

        resp = self.execute('DescribeVpnConnections', {})
        vpns = [tools.update_dict(
                    vpn, {'customerGatewayConfiguration': 'DONTCARE'})
                for vpn in (fakes.EC2_VPN_CONNECTION_1,
                            fakes.EC2_VPN_CONNECTION_2)]
        self.assertThat(
            resp,
            matchers.DictMatches(
                {'vpnConnectionSet': vpns},
                orderless_lists=True))
        for vpn in (fakes.EC2_VPN_CONNECTION_1, fakes.EC2_VPN_CONNECTION_2):
            config = next(v['customerGatewayConfiguration']
                          for v in resp['vpnConnectionSet']
                          if v['vpnConnectionId'] == vpn['vpnConnectionId'])
            self.assertThat(
                config.encode(),
                matchers.XMLMatches(
                    vpn['customerGatewayConfiguration'].encode(),
                    orderless_sequence=True))
            self.assertTrue(config.startswith(
                '<?xml version=\'1.0\' encoding=\'UTF-8\'?>'))
        self.neutron.list_ikepolicies.assert_called_once_with(
            tenant_id=fakes.ID_OS_PROJECT)
        self.neutron.list_ipsecpolicies.assert_called_once_with(
            tenant_id=fakes.ID_OS_PROJECT)
        self.neutron.list_ipsec_site_connections.assert_called_once_with(
            tenant_id=fakes.ID_OS_PROJECT)
        self.neutron.list_routers.assert_called_once_with(
            tenant_id=fakes.ID_OS_PROJECT)

        resp = self.execute(
            'DescribeVpnConnections',
            {'VpnConnectionId.1': fakes.ID_EC2_VPN_CONNECTION_1})
        self.assertThat(
            resp,
            matchers.DictMatches(
                {'vpnConnectionSet': [vpns[0]]},
                orderless_lists=True))

        self.check_filtering(
            'DescribeVpnConnections', 'vpnConnectionSet',
            [('customer-gateway-configuration',
              '*' + fakes.PRE_SHARED_KEY_1 + '*'),
             ('customer-gateway-id', fakes.ID_EC2_CUSTOMER_GATEWAY_1),
             ('state', 'available'),
             ('option.static-routes-only', True),
             ('route.destination-cidr-block', fakes.CIDR_VPN_2_PROPAGATED_1),
             ('type', 'ipsec.1'),
             ('vpn-connection-id', fakes.ID_EC2_VPN_CONNECTION_1),
             ('vpn-gateway-id', fakes.ID_EC2_VPN_GATEWAY_1)])

        self.check_tag_support(
            'DescribeVpnConnections', 'vpnConnectionSet',
            fakes.ID_EC2_VPN_CONNECTION_1, 'vpnConnectionId')

    def test_format_vpn_connection(self):
        db_vpn_connection_1 = tools.update_dict(fakes.DB_VPN_CONNECTION_1,
                                                {'cidrs': []})
        ec2_vpn_connection_1 = tools.patch_dict(
            fakes.EC2_VPN_CONNECTION_1,
            {'routes': [], 'vgwTelemetry': []},
            ('customerGatewayConfiguration',))
        formatted = vpn_connection_api._format_vpn_connection(
            db_vpn_connection_1,
            {fakes.ID_EC2_CUSTOMER_GATEWAY_1: fakes.DB_CUSTOMER_GATEWAY_1},
            {}, {}, {}, {})
        formatted.pop('customerGatewayConfiguration')
        self.assertThat(ec2_vpn_connection_1, matchers.DictMatches(formatted))

    def test_format_customer_config(self):
        ikepolicy = {
            'auth_algorithm': 'sha1-fake',
            'encryption_algorithm': '3des',
            'lifetime': {'value': 1111},
            'pfs': 'group5',
            'phase1_negotiation_mode': 'main-fake',
        }
        ipsecpolicy = {
            'transform_protocol': 'ah-esp',
            'auth_algorithm': 'sha1-fake',
            'encryption_algorithm': 'aes-256',
            'lifetime': {'value': 2222},
            'pfs': 'group14',
            'encapsulation_mode': 'transport',
        }
        ipsec_site_connection = {
            'peer_address': '1.2.3.4',
            'psk': 'password',
            'mtu': 1400,
        }
        conf = vpn_connection_api._format_customer_config(
            fakes.DB_VPN_CONNECTION_1,
            {fakes.ID_EC2_CUSTOMER_GATEWAY_1: fakes.DB_CUSTOMER_GATEWAY_1},
            {fakes.ID_OS_IKEPOLICY_1: ikepolicy},
            {fakes.ID_OS_IPSECPOLICY_1: ipsecpolicy},
            {fakes.ID_OS_IPSEC_SITE_CONNECTION_2: ipsec_site_connection},
            {fakes.ID_EC2_VPN_GATEWAY_1: '5.6.7.8'})

        self.assertThat(
            {'ipsec_tunnel': {
                'customer_gateway': {
                    'tunnel_outside_address': {'ip_address': '1.2.3.4'}},
                'vpn_gateway': {
                    'tunnel_outside_address': {'ip_address': '5.6.7.8'}},
                'ike': {'authentication_protocol': 'sha1-fake',
                        'encryption_protocol': '3des',
                        'lifetime': 1111,
                        'perfect_forward_secrecy': 'group5',
                        'mode': 'main-fake',
                        'pre_shared_key': 'password'},
                'ipsec': {'protocol': 'ah-esp',
                          'authentication_protocol': 'sha1-fake',
                          'encryption_protocol': 'aes-256',
                          'lifetime': 2222,
                          'perfect_forward_secrecy': 'group14',
                          'mode': 'transport',
                          'tcp_mss_adjustment': 1400 - 40}}},
            matchers.IsSubDictOf(conf))

    def test_stop_vpn_connection(self):
        # delete several connections
        os_conn_ids = [fakes.random_os_id() for _x in range(3)]
        fake_conn = {
            'os_ipsec_site_connections': {
                fakes.random_ec2_id('subnet'): conn_id
                for conn_id in os_conn_ids}}
        vpn_connection_api._stop_vpn_connection(self.neutron, fake_conn)
        self.assertEqual(
            3, self.neutron.delete_ipsec_site_connection.call_count)
        for conn_id in os_conn_ids:
            self.neutron.delete_ipsec_site_connection.assert_any_call(conn_id)

        # delete several connections with exception suppressing
        self.neutron.reset_mock()
        self.neutron.delete_ipsec_site_connection.side_effect = [
            None, neutron_exception.NotFound(), None]
        vpn_connection_api._stop_vpn_connection(self.neutron, fake_conn)
        self.assertEqual(
            3, self.neutron.delete_ipsec_site_connection.call_count)

    @mock.patch('ec2api.api.vpn_connection._stop_vpn_connection',
                new_callable=tools.CopyingMock)
    def test_stop_gateway_vpn_connections(self, stop_vpn_connection):
        context = base.create_context()
        cleaner = common.OnCrashCleaner()
        vpn_connection_3 = tools.update_dict(
            fakes.DB_VPN_CONNECTION_1,
            {'id': fakes.random_ec2_id('vpn'),
             'os_ipsec_site_connections': {}})

        self.set_mock_db_items(fakes.DB_VPN_CONNECTION_1, vpn_connection_3,
                               fakes.DB_VPN_CONNECTION_2)
        vpn_connection_api._stop_gateway_vpn_connections(
            context, self.neutron, cleaner, fakes.DB_VPN_GATEWAY_1)
        self.assertEqual(2, stop_vpn_connection.call_count)
        stop_vpn_connection.assert_any_call(
            self.neutron, fakes.DB_VPN_CONNECTION_1)
        stop_vpn_connection.assert_any_call(
            self.neutron, vpn_connection_3)
        self.assertEqual(2, self.db_api.update_item.call_count)
        self.db_api.update_item.assert_any_call(
            mock.ANY, tools.update_dict(fakes.DB_VPN_CONNECTION_1,
                                        {'os_ipsec_site_connections': {}}))
        self.db_api.update_item.assert_any_call(
            mock.ANY, vpn_connection_3)

        self.db_api.reset_mock()
        self.neutron.reset_mock()
        stop_vpn_connection.reset_mock()
        self.set_mock_db_items(fakes.DB_VPN_CONNECTION_1)
        try:
            with common.OnCrashCleaner() as cleaner:
                vpn_connection_api._stop_gateway_vpn_connections(
                    context, self.neutron, cleaner, fakes.DB_VPN_GATEWAY_1)
                raise Exception('fake-exception')
        except Exception as ex:
            if str(ex) != 'fake-exception':
                raise
        self.db_api.update_item.assert_called_with(
            mock.ANY, fakes.DB_VPN_CONNECTION_1)

    @mock.patch('ec2api.api.vpn_connection._reset_vpn_connections')
    def test_update_vpn_routes(self, reset_vpn_connections):
        context = base.create_context()
        cleaner = common.OnCrashCleaner()

        self.set_mock_db_items()
        vpn_connection_api._update_vpn_routes(
            context, self.neutron, cleaner,
            fakes.DB_ROUTE_TABLE_1, [fakes.DB_SUBNET_1])
        self.assertFalse(reset_vpn_connections.called)

        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_1)
        vpn_connection_api._update_vpn_routes(
            context, self.neutron, cleaner,
            fakes.DB_ROUTE_TABLE_1, [fakes.DB_SUBNET_1])
        reset_vpn_connections.assert_called_once_with(
            context, self.neutron, cleaner, fakes.DB_VPN_GATEWAY_1,
            route_tables=[fakes.DB_ROUTE_TABLE_1], subnets=[fakes.DB_SUBNET_1])

    @mock.patch('ec2api.api.vpn_connection._delete_subnet_vpn')
    @mock.patch('ec2api.api.vpn_connection._set_subnet_vpn')
    @mock.patch('ec2api.api.vpn_connection._get_route_table_vpn_cidrs',
                wraps=vpn_connection_api._get_route_table_vpn_cidrs)
    def test_reset_vpn_connections(self, get_route_table_vpn_cidrs,
                                   set_subnet_vpn, delete_subnet_vpn):
        context = base.create_context()
        cleaner = common.OnCrashCleaner()

        vpn_gateway_3 = {'id': fakes.random_ec2_id('vpn'),
                         'os_id': None,
                         'vpc_id': None}
        vpn_connection_api._reset_vpn_connections(
            context, self.neutron, cleaner, vpn_gateway_3)
        self.assertEqual(0, len(self.db_api.mock_calls))
        self.assertFalse(get_route_table_vpn_cidrs.called)
        self.assertFalse(set_subnet_vpn.called)
        self.assertFalse(delete_subnet_vpn.called)

        customer_gateway_3 = {'id': fakes.random_ec2_id('cgw')}
        subnet_3 = {'id': fakes.random_ec2_id('subnet'),
                    'vpc_id': fakes.ID_EC2_VPC_2}
        vpn_connection_3 = {'id': fakes.random_ec2_id('vpn'),
                            'vpn_gateway_id': fakes.ID_EC2_VPN_GATEWAY_1,
                            'customer_gateway_id': customer_gateway_3['id'],
                            'cidrs': []}
        self.set_mock_db_items(
            fakes.DB_VPC_1, fakes.DB_VPC_2,
            fakes.DB_CUSTOMER_GATEWAY_1, fakes.DB_CUSTOMER_GATEWAY_2,
            customer_gateway_3,
            fakes.DB_SUBNET_1, fakes.DB_SUBNET_2, subnet_3,
            fakes.DB_ROUTE_TABLE_1, fakes.DB_ROUTE_TABLE_2,
            fakes.DB_ROUTE_TABLE_3,
            fakes.DB_VPN_CONNECTION_1, fakes.DB_VPN_CONNECTION_2,
            vpn_connection_3)

        # common case
        vpn_connection_api._reset_vpn_connections(
            context, self.neutron, cleaner, fakes.DB_VPN_GATEWAY_1)
        self.assertEqual(2, set_subnet_vpn.call_count)
        set_subnet_vpn.assert_any_call(
            context, self.neutron, cleaner, fakes.DB_SUBNET_2,
            fakes.DB_VPN_CONNECTION_1, fakes.DB_CUSTOMER_GATEWAY_1,
            [fakes.CIDR_VPN_1_STATIC])
        set_subnet_vpn.assert_any_call(
            context, self.neutron, cleaner, fakes.DB_SUBNET_2,
            vpn_connection_3, customer_gateway_3,
            [fakes.CIDR_VPN_1_STATIC])
        self.assertEqual(2, delete_subnet_vpn.call_count)
        delete_subnet_vpn.assert_any_call(
            context, self.neutron, cleaner, fakes.DB_SUBNET_1,
            fakes.DB_VPN_CONNECTION_1)
        delete_subnet_vpn.assert_any_call(
            context, self.neutron, cleaner, fakes.DB_SUBNET_1,
            vpn_connection_3)
        self.assertEqual(2, get_route_table_vpn_cidrs.call_count)
        get_route_table_vpn_cidrs.assert_any_call(
            fakes.DB_ROUTE_TABLE_1, fakes.DB_VPN_GATEWAY_1,
            [fakes.DB_VPN_CONNECTION_1, vpn_connection_3])
        get_route_table_vpn_cidrs.assert_any_call(
            fakes.DB_ROUTE_TABLE_3, fakes.DB_VPN_GATEWAY_1,
            [fakes.DB_VPN_CONNECTION_1, vpn_connection_3])

        # reset for the vpn connection
        set_subnet_vpn.reset_mock()
        delete_subnet_vpn.reset_mock()
        self.db_api.reset_mock()
        get_route_table_vpn_cidrs.reset_mock()
        vpn_connection_api._reset_vpn_connections(
            context, self.neutron, cleaner, fakes.DB_VPN_GATEWAY_1,
            vpn_connections=[fakes.DB_VPN_CONNECTION_1])
        self.assertEqual(1, set_subnet_vpn.call_count)
        self.assertEqual(1, delete_subnet_vpn.call_count)
        self.assertNotIn(mock.call(mock.ANY, 'vpn'),
                         self.db_api.get_items.mock_calls)

        # reset for the subnet list
        set_subnet_vpn.reset_mock()
        delete_subnet_vpn.reset_mock()
        self.db_api.reset_mock()
        get_route_table_vpn_cidrs.reset_mock()
        vpn_connection_api._reset_vpn_connections(
            context, self.neutron, cleaner, fakes.DB_VPN_GATEWAY_1,
            subnets=[fakes.DB_SUBNET_1])
        self.assertFalse(set_subnet_vpn.called)
        self.assertEqual(2, delete_subnet_vpn.call_count)
        self.assertNotIn(mock.call(mock.ANY, 'subnets'),
                         self.db_api.get_items.mock_calls)

        # reset for the subnet list and the route table
        set_subnet_vpn.reset_mock()
        delete_subnet_vpn.reset_mock()
        self.db_api.reset_mock()
        get_route_table_vpn_cidrs.reset_mock()
        vpn_connection_api._reset_vpn_connections(
            context, self.neutron, cleaner, fakes.DB_VPN_GATEWAY_1,
            subnets=[fakes.DB_SUBNET_2], route_tables=[fakes.DB_ROUTE_TABLE_3])
        self.assertEqual(2, set_subnet_vpn.call_count)
        self.assertFalse(delete_subnet_vpn.called)
        self.assertNotIn(mock.call(mock.ANY, 'subnets'),
                         self.db_api.get_items.mock_calls)
        self.assertNotIn(mock.call(mock.ANY, 'rtb'),
                         self.db_api.get_items.mock_calls)

    def test_set_subnet_vpn(self):
        context = base.create_context()
        cleaner = common.OnCrashCleaner()
        cidrs = [fakes.CIDR_VPN_1_STATIC, fakes.CIDR_VPN_1_PROPAGATED_1]

        # create ipsec site connection case
        id_os_connection = fakes.random_os_id()
        os_connection = {
            'vpnservice_id': fakes.ID_OS_VPNSERVICE_1,
            'ikepolicy_id': fakes.ID_OS_IKEPOLICY_1,
            'ipsecpolicy_id': fakes.ID_OS_IPSECPOLICY_1,
            'peer_address': fakes.IP_CUSTOMER_GATEWAY_ADDRESS_1,
            'peer_cidrs': cidrs,
            'psk': fakes.PRE_SHARED_KEY_1,
            'name': (fakes.ID_EC2_VPN_CONNECTION_1 + '/' +
                     fakes.ID_EC2_SUBNET_1),
            'peer_id': fakes.IP_CUSTOMER_GATEWAY_ADDRESS_1,
            'mtu': 1427,
            'initiator': 'response-only',
        }
        self.neutron.create_ipsec_site_connection.side_effect = (
            tools.get_neutron_create('ipsec_site_connection',
                                     id_os_connection))
        vpn_connection_api._set_subnet_vpn(
            context, self.neutron, cleaner, fakes.DB_SUBNET_1,
            copy.deepcopy(fakes.DB_VPN_CONNECTION_1),
            fakes.DB_CUSTOMER_GATEWAY_1, cidrs)

        self.neutron.create_ipsec_site_connection.assert_called_once_with(
            {'ipsec_site_connection': os_connection})
        vpn_connection_1 = copy.deepcopy(fakes.DB_VPN_CONNECTION_1)
        (vpn_connection_1['os_ipsec_site_connections']
         [fakes.ID_EC2_SUBNET_1]) = id_os_connection
        self.db_api.update_item.assert_called_once_with(
            context, vpn_connection_1)

        # update ipsec site connection case
        self.db_api.reset_mock()
        self.neutron.reset_mock()
        vpn_connection_api._set_subnet_vpn(
            context, self.neutron, cleaner, fakes.DB_SUBNET_2,
            fakes.DB_VPN_CONNECTION_1, fakes.DB_CUSTOMER_GATEWAY_1, cidrs)
        self.neutron.update_ipsec_site_connection.assert_called_once_with(
            fakes.ID_OS_IPSEC_SITE_CONNECTION_2,
            {'ipsec_site_connection': {'peer_cidrs': cidrs}})
        self.assertFalse(self.neutron.create_ipsec_site_connection.called)
        self.assertFalse(self.db_api.update_item.called)

        # rollback creating of ipsec site connection case
        self.db_api.reset_mock()
        self.neutron.reset_mock()
        try:
            with common.OnCrashCleaner() as cleaner:
                vpn_connection_api._set_subnet_vpn(
                    context, self.neutron, cleaner, fakes.DB_SUBNET_1,
                    copy.deepcopy(fakes.DB_VPN_CONNECTION_1),
                    fakes.DB_CUSTOMER_GATEWAY_1, cidrs)
                raise Exception('fake-exception')
        except Exception as ex:
            if str(ex) != 'fake-exception':
                raise
        self.neutron.delete_ipsec_site_connection.assert_called_once_with(
            id_os_connection)
        self.db_api.update_item.assert_called_with(
            mock.ANY, fakes.DB_VPN_CONNECTION_1)

        # rollback updating of ipsec site connection case
        self.db_api.reset_mock()
        self.neutron.reset_mock()
        try:
            with common.OnCrashCleaner() as cleaner:
                vpn_connection_api._set_subnet_vpn(
                    context, self.neutron, cleaner, fakes.DB_SUBNET_2,
                    fakes.DB_VPN_CONNECTION_1, fakes.DB_CUSTOMER_GATEWAY_1,
                    cidrs)
                raise Exception('fake-exception')
        except Exception as ex:
            if str(ex) != 'fake-exception':
                raise
        self.assertFalse(self.neutron.delete_ipsec_site_connection.called)
        self.assertFalse(self.db_api.update_item.called)

    def test_delete_subnet_vpn(self):
        context = base.create_context()
        cleaner = common.OnCrashCleaner()

        # subnet is not connected to the vpn
        vpn_connection_api._delete_subnet_vpn(
            context, self.neutron, cleaner, fakes.DB_SUBNET_1,
            fakes.DB_VPN_CONNECTION_1)
        self.assertFalse(self.db_api.update_item.called)
        self.assertFalse(self.neutron.delete_ipsec_site_connection.called)

        # delete subnet vpn connection
        vpn_connection_api._delete_subnet_vpn(
            context, self.neutron, cleaner, fakes.DB_SUBNET_2,
            copy.deepcopy(fakes.DB_VPN_CONNECTION_1))
        self.db_api.update_item.assert_called_once_with(
            mock.ANY, tools.update_dict(fakes.DB_VPN_CONNECTION_1,
                                        {'os_ipsec_site_connections': {}}))
        self.neutron.delete_ipsec_site_connection.assert_called_once_with(
            fakes.ID_OS_IPSEC_SITE_CONNECTION_2)

        # delete subnet vpn connection, leave connections of other subnets
        self.db_api.reset_mock()
        self.neutron.reset_mock()
        id_os_connection = fakes.random_os_id()
        vpn_connection_1 = copy.deepcopy(fakes.DB_VPN_CONNECTION_1)
        (vpn_connection_1['os_ipsec_site_connections']
         [fakes.ID_EC2_SUBNET_1]) = id_os_connection
        vpn_connection_api._delete_subnet_vpn(
            context, self.neutron, cleaner, fakes.DB_SUBNET_1,
            vpn_connection_1)
        self.db_api.update_item.assert_called_once_with(
            mock.ANY, fakes.DB_VPN_CONNECTION_1)
        self.neutron.delete_ipsec_site_connection.assert_called_once_with(
            id_os_connection)

        # rollback of deleting subnet vpn connection
        self.db_api.reset_mock()
        self.neutron.reset_mock()
        try:
            with common.OnCrashCleaner() as cleaner:
                vpn_connection_api._delete_subnet_vpn(
                    context, self.neutron, cleaner, fakes.DB_SUBNET_2,
                    copy.deepcopy(fakes.DB_VPN_CONNECTION_1))
                raise Exception('fake-exception')
        except Exception as ex:
            if str(ex) != 'fake-exception':
                raise
        self.db_api.update_item.assert_called_with(
            mock.ANY, fakes.DB_VPN_CONNECTION_1)
        self.assertFalse(self.neutron.create_ipsec_site_connection.called)

    def test_get_route_table_vpn_cidrs(self):
        route_table_1 = copy.deepcopy(fakes.DB_ROUTE_TABLE_1)
        vpn_connection_1 = tools.update_dict(
            fakes.DB_VPN_CONNECTION_1, {'cidrs': []})
        vpn_connection_2 = tools.update_dict(
            vpn_connection_1, {'id': fakes.ID_EC2_VPN_CONNECTION_2})

        self.assertThat(
            vpn_connection_api._get_route_table_vpn_cidrs(
                route_table_1, fakes.DB_VPN_GATEWAY_1, []),
            matchers.DictMatches({}))

        self.assertThat(
            vpn_connection_api._get_route_table_vpn_cidrs(
                route_table_1, fakes.DB_VPN_GATEWAY_1,
                [vpn_connection_1, vpn_connection_2]),
            matchers.DictMatches({}))

        route_table_1['propagating_gateways'] = [fakes.ID_EC2_VPN_GATEWAY_1,
                                                 fakes.ID_EC2_VPN_GATEWAY_2]
        self.assertThat(
            vpn_connection_api._get_route_table_vpn_cidrs(
                route_table_1, fakes.DB_VPN_GATEWAY_1,
                [vpn_connection_1, vpn_connection_2]),
            matchers.DictMatches({}))

        vpn_connection_1['cidrs'] = ['cidr_1']
        self.assertThat(
            vpn_connection_api._get_route_table_vpn_cidrs(
                route_table_1, fakes.DB_VPN_GATEWAY_1,
                [vpn_connection_1, vpn_connection_2]),
            matchers.DictMatches({fakes.ID_EC2_VPN_CONNECTION_1: ['cidr_1']}))

        vpn_connection_2['cidrs'] = ['cidr_1', 'cidr_2']
        self.assertThat(
            vpn_connection_api._get_route_table_vpn_cidrs(
                route_table_1, fakes.DB_VPN_GATEWAY_1,
                [vpn_connection_1, vpn_connection_2]),
            matchers.DictMatches(
                {fakes.ID_EC2_VPN_CONNECTION_1: ['cidr_1'],
                 fakes.ID_EC2_VPN_CONNECTION_2: ['cidr_1', 'cidr_2']},
                orderless_lists=True))

        route_table_1['routes'] = [
            {'destination_cidr_block': 'fake_1',
             'network_interface_id': fakes.ID_EC2_NETWORK_INTERFACE_1},
            {'destination_cidr_block': 'fake_2',
             'gateway_id': None},
            {'destination_cidr_block': 'fake_3',
             'gateway_id': fakes.ID_EC2_IGW_1},
            {'destination_cidr_block': 'cidr_3',
             'gateway_id': fakes.ID_EC2_VPN_GATEWAY_1},
            {'destination_cidr_block': 'cidr_4',
             'gateway_id': fakes.ID_EC2_VPN_GATEWAY_1},
            {'destination_cidr_block': 'fake_4',
             'gateway_id': fakes.ID_EC2_VPN_GATEWAY_2}]

        self.assertThat(
            vpn_connection_api._get_route_table_vpn_cidrs(
                route_table_1, fakes.DB_VPN_GATEWAY_1,
                [vpn_connection_1, vpn_connection_2]),
            matchers.DictMatches(
                {fakes.ID_EC2_VPN_CONNECTION_1: ['cidr_1', 'cidr_3', 'cidr_4'],
                 fakes.ID_EC2_VPN_CONNECTION_2: ['cidr_1', 'cidr_2',
                                                 'cidr_3', 'cidr_4']},
                orderless_lists=True))

        route_table_1['propagating_gateways'] = [fakes.ID_EC2_VPN_GATEWAY_2]
        self.assertThat(
            vpn_connection_api._get_route_table_vpn_cidrs(
                route_table_1, fakes.DB_VPN_GATEWAY_1,
                [vpn_connection_1, vpn_connection_2]),
            matchers.DictMatches(
                {fakes.ID_EC2_VPN_CONNECTION_1: ['cidr_3', 'cidr_4'],
                 fakes.ID_EC2_VPN_CONNECTION_2: ['cidr_3', 'cidr_4']},
                orderless_lists=True))



# Copyright 2014
# The Cloudscaling Group, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from unittest import mock

import fixtures
from neutronclient.common import exceptions as neutron_exception

from ec2api.api import common
from ec2api.api import vpn_gateway as vpn_gateway_api
from ec2api.tests.unit import base
from ec2api.tests.unit import fakes
from ec2api.tests.unit import matchers
from ec2api.tests.unit import tools


class VpnGatewayTestCase(base.ApiTestCase):

    def setUp(self):
        super(VpnGatewayTestCase, self).setUp()
        self.DB_VPN_GATEWAY_2_ATTACHED = tools.update_dict(
            fakes.DB_VPN_GATEWAY_2, {'vpc_id': fakes.ID_EC2_VPC_2})
        self.DB_VPN_GATEWAY_1_DETACHED = tools.update_dict(
            fakes.DB_VPN_GATEWAY_1, {'vpc_id': None})
        self.DB_SUBNET_1_NO_VPN = tools.purge_dict(
            fakes.DB_SUBNET_1, ('os_vpnservice_id',))

    def test_create_vpn_gateway(self):
        self.db_api.add_item.side_effect = (
            tools.get_db_api_add_item(fakes.ID_EC2_VPN_GATEWAY_2))

        resp = self.execute('CreateVpnGateway',
                            {'Type': 'ipsec.1'})
        self.assertEqual({'vpnGateway': fakes.EC2_VPN_GATEWAY_2}, resp)
        self.db_api.add_item.assert_called_once_with(mock.ANY, 'vgw', {})

    def test_attach_vpn_gateway(self):
        self.configure(external_network=fakes.NAME_OS_PUBLIC_NETWORK)
        subnet_2 = tools.patch_dict(fakes.DB_SUBNET_2,
                                    {'vpc_id': fakes.ID_EC2_VPC_2},
                                    ('os_vpnservice_id',))
        self.set_mock_db_items(
            fakes.DB_VPN_GATEWAY_1, fakes.DB_VPN_GATEWAY_2,
            fakes.DB_VPC_2, fakes.DB_IGW_1, fakes.DB_IGW_2,
            fakes.DB_SUBNET_1, subnet_2)
        subnet_2_updated = tools.update_dict(
            subnet_2, {'os_vpnservice_id': fakes.ID_OS_VPNSERVICE_2})
        os_vpnservice_2 = tools.patch_dict(fakes.OS_VPNSERVICE_2,
                                           {'router_id': fakes.ID_OS_ROUTER_2},
                                           ('id',))
        self.neutron.list_networks.return_value = (
            {'networks': [{'id': fakes.ID_OS_PUBLIC_NETWORK}]})
        self.neutron.create_vpnservice.side_effect = tools.get_neutron_create(
            'vpnservice', fakes.ID_OS_VPNSERVICE_2)

        def do_check():
            resp = self.execute('AttachVpnGateway',
                                {'VpcId': fakes.ID_EC2_VPC_2,
                                 'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_2})
            self.assertEqual({'attachment': {'state': 'attached',
                                             'vpcId': fakes.ID_EC2_VPC_2}},
                             resp)
            self.assertEqual(2, self.db_api.update_item.call_count)
            self.db_api.update_item.assert_has_calls(
                [mock.call(mock.ANY, self.DB_VPN_GATEWAY_2_ATTACHED),
                 mock.call(mock.ANY, subnet_2_updated)])
            self.neutron.create_vpnservice.assert_called_once_with(
                {'vpnservice': os_vpnservice_2})

        do_check()
        self.neutron.add_gateway_router.assert_called_once_with(
            fakes.ID_OS_ROUTER_2,
            {'network_id': fakes.ID_OS_PUBLIC_NETWORK})
        self.neutron.list_networks.assert_called_once_with(
            **{'router:external': True,
               'name': fakes.NAME_OS_PUBLIC_NETWORK})

        # Internet gateway is already attached
        self.db_api.reset_mock()
        self.neutron.reset_mock()
        igw_2 = tools.update_dict(fakes.DB_IGW_2,
                                  {'vpc_id': fakes.ID_EC2_VPC_2})
        self.add_mock_db_items(igw_2)

        do_check()
        self.neutron.add_gateway_router.assert_not_called()

    def test_attach_vpn_gateway_idempotent(self):
        self.configure(external_network=fakes.NAME_OS_PUBLIC_NETWORK)
        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_1, fakes.DB_VPC_1)

        resp = self.execute('AttachVpnGateway',
                            {'VpcId': fakes.ID_EC2_VPC_1,
                             'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_1})
        self.assertEqual({'attachment': {'state': 'attached',
                                         'vpcId': fakes.ID_EC2_VPC_1}},
                         resp)
        self.assertFalse(self.db_api.update_item.called)
        self.assertFalse(self.neutron.add_gateway_router.called)
        self.assertFalse(self.neutron.create_vpnservice.called)

    def test_attach_vpn_gateway_invalid_parameters(self):
        def do_check(error_code):
            self.assert_execution_error(
                error_code, 'AttachVpnGateway',
                {'VpcId': fakes.ID_EC2_VPC_2,
                 'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_2})

            self.assertFalse(self.db_api.update_item.called)
            self.db_api.reset_mock()

        self.set_mock_db_items(fakes.DB_VPC_2)
        do_check('InvalidVpnGatewayID.NotFound')

        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_2)
        do_check('InvalidVpcID.NotFound')

        self.set_mock_db_items(
            tools.update_dict(fakes.DB_VPN_GATEWAY_2,
                              {'vpc_id': fakes.ID_EC2_VPC_1}),
            fakes.DB_VPC_2)
        do_check('VpnGatewayAttachmentLimitExceeded')

        self.set_mock_db_items(
            fakes.DB_VPN_GATEWAY_2, fakes.DB_VPC_2,
            tools.update_dict(fakes.DB_VPN_GATEWAY_1,
                              {'vpc_id': fakes.ID_EC2_VPC_2}))
        do_check('InvalidVpcState')

    @tools.screen_unexpected_exception_logs
    def test_attach_vpn_gateway_rollback(self):
        self.configure(external_network=fakes.NAME_OS_PUBLIC_NETWORK)
        subnet_2 = tools.patch_dict(fakes.DB_SUBNET_2,
                                    {'vpc_id': fakes.ID_EC2_VPC_2},
                                    ('os_vpnservice_id',))
        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_2, fakes.DB_VPC_2,
                               subnet_2)
        self.neutron.list_networks.return_value = (
            {'networks': [{'id': fakes.ID_OS_PUBLIC_NETWORK}]})
        self.neutron.create_vpnservice.side_effect = Exception()

        self.assert_execution_error(
            self.ANY_EXECUTE_ERROR, 'AttachVpnGateway',
            {'VpcId': fakes.ID_EC2_VPC_2,
             'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_2})

        self.db_api.update_item.assert_any_call(
            mock.ANY, fakes.DB_VPN_GATEWAY_2)
        self.neutron.remove_gateway_router.assert_called_once_with(
            fakes.ID_OS_ROUTER_2)

    def test_detach_vpn_gateway(self):
        self.set_mock_db_items(
            fakes.DB_VPN_GATEWAY_1, fakes.DB_VPC_1, fakes.DB_VPN_CONNECTION_1,
            fakes.DB_SUBNET_1,
            tools.update_dict(fakes.DB_SUBNET_2,
                              {'vpc_id': fakes.ID_EC2_VPC_2}))

        def do_check():
            resp = self.execute(
                'DetachVpnGateway',
                {'VpcId': fakes.ID_EC2_VPC_1,
                 'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_1})
            self.assertEqual({'return': True}, resp)
            self.assertEqual(3, self.db_api.update_item.call_count)
            self.db_api.update_item.assert_has_calls(
                [mock.call(mock.ANY, self.DB_VPN_GATEWAY_1_DETACHED),
                 mock.call(mock.ANY,
                           tools.update_dict(
                               fakes.DB_VPN_CONNECTION_1,
                               {'os_ipsec_site_connections': {}})),
                 mock.call(mock.ANY, self.DB_SUBNET_1_NO_VPN)])
            self.neutron.delete_vpnservice.assert_called_once_with(
                fakes.ID_OS_VPNSERVICE_1)
            self.neutron.delete_ipsec_site_connection.assert_called_once_with(
                fakes.ID_OS_IPSEC_SITE_CONNECTION_2)

        do_check()
        self.neutron.remove_gateway_router.assert_called_once_with(
            fakes.ID_OS_ROUTER_1)

        # Internet gateway is still attached
        self.db_api.reset_mock()
        self.neutron.reset_mock()
        self.add_mock_db_items(fakes.DB_IGW_1)

        do_check()
        self.neutron.remove_gateway_router.assert_not_called()

    def test_detach_vpn_gateway_invalid_parameters(self):
        def do_check(error_code):
            self.assert_execution_error(
                error_code, 'DetachVpnGateway',
                {'VpcId': fakes.ID_EC2_VPC_1,
                 'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_2})

            self.assertEqual(0, self.neutron.remove_gateway_router.call_count)
            self.assertEqual(0, self.db_api.update_item.call_count)

            self.neutron.reset_mock()
            self.db_api.reset_mock()

        self.set_mock_db_items()
        do_check('InvalidVpnGatewayID.NotFound')

        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_2)
        do_check('InvalidVpnGatewayAttachment.NotFound')

        self.set_mock_db_items(self.DB_VPN_GATEWAY_2_ATTACHED)
        do_check('InvalidVpnGatewayAttachment.NotFound')

    def test_detach_vpn_gateway_no_router(self):
        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_1, fakes.DB_VPC_1)
        self.neutron.remove_gateway_router.side_effect = (
            neutron_exception.NotFound)

        resp = self.execute(
            'DetachVpnGateway',
            {'VpcId': fakes.ID_EC2_VPC_1,
             'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_1})

        self.assertEqual(True, resp['return'])
        self.neutron.remove_gateway_router.assert_called_once_with(
            fakes.ID_OS_ROUTER_1)

    @tools.screen_unexpected_exception_logs
    def test_detach_vpn_gateway_rollback(self):
        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_1, fakes.DB_VPC_1,
                               fakes.DB_SUBNET_1)
        self.neutron.remove_gateway_router.side_effect = Exception()

        self.assert_execution_error(
            self.ANY_EXECUTE_ERROR, 'DetachVpnGateway',
            {'VpcId': fakes.ID_EC2_VPC_1,
             'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_1})

        self.db_api.update_item.assert_has_calls(
            [mock.call(mock.ANY, fakes.DB_SUBNET_1),
             mock.call(mock.ANY, fakes.DB_VPN_GATEWAY_1)])

    def test_delete_vpn_gateway(self):
        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_2)

        resp = self.execute(
            'DeleteVpnGateway',
            {'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_2})

        self.assertEqual({'return': True}, resp)
        self.db_api.delete_item.assert_called_once_with(
            mock.ANY, fakes.ID_EC2_VPN_GATEWAY_2)

    def test_delete_vpn_gateway_invalid_parameters(self):
        self.set_mock_db_items()
        self.assert_execution_error(
            'InvalidVpnGatewayID.NotFound', 'DeleteVpnGateway',
            {'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_1})

        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_1)
        self.assert_execution_error(
            'IncorrectState', 'DeleteVpnGateway',
            {'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_1})

        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_2,
                               fakes.DB_VPN_CONNECTION_2)
        self.assert_execution_error(
            'IncorrectState', 'DeleteVpnGateway',
            {'VpnGatewayId': fakes.ID_EC2_VPN_GATEWAY_2})

    def test_describe_vpn_gateways(self):
        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_1, fakes.DB_VPN_GATEWAY_2)

        resp = self.execute('DescribeVpnGateways', {})
        self.assertThat(resp['vpnGatewaySet'],
                        matchers.ListMatches([fakes.EC2_VPN_GATEWAY_1,
                                              fakes.EC2_VPN_GATEWAY_2]))

        resp = self.execute('DescribeVpnGateways',
                            {'VpnGatewayId.1': fakes.ID_EC2_VPN_GATEWAY_2})
        self.assertThat(resp['vpnGatewaySet'],
                        matchers.ListMatches([fakes.EC2_VPN_GATEWAY_2]))
        self.db_api.get_items_by_ids.assert_called_once_with(
            mock.ANY, set([fakes.ID_EC2_VPN_GATEWAY_2]))

        self.check_filtering(
            'DescribeVpnGateways', 'vpnGatewaySet',
            [('attachment.state', 'attached'),
             ('attachment.vpc-id', fakes.ID_EC2_VPC_1),
             ('state', 'available'),
             ('type', 'ipsec.1'),
             ('vpn-gateway-id', fakes.ID_EC2_VPN_GATEWAY_2)])
        self.check_tag_support(
            'DescribeVpnGateways', 'vpnGatewaySet',
            fakes.ID_EC2_VPN_GATEWAY_2, 'vpnGatewayId')

    @mock.patch('ec2api.api.vpn_connection._reset_vpn_connections')
    @mock.patch('ec2api.api.vpn_gateway._create_subnet_vpnservice')
    def test_start_vpn_in_subnet(self, create_subnet_vpnservice,
                                 reset_vpn_connection):
        context = base.create_context()
        cleaner = common.OnCrashCleaner()
        mock_manager = mock.Mock()
        mock_manager.attach_mock(create_subnet_vpnservice,
                                 'create_subnet_vpnservice')
        mock_manager.attach_mock(reset_vpn_connection, 'reset_vpn_connection')

        self.set_mock_db_items(fakes.DB_VPN_GATEWAY_1, fakes.DB_VPN_GATEWAY_2)
        vpn_gateway_api._start_vpn_in_subnet(
            context, self.neutron, cleaner, copy.deepcopy(fakes.DB_SUBNET_1),
            fakes.DB_VPC_1, fakes.DB_ROUTE_TABLE_1)
        mock_manager.assert_has_calls([
            mock.call.create_subnet_vpnservice(
                context, self.neutron, cleaner,
                fakes.DB_SUBNET_1, fakes.DB_VPC_1),
            mock.call.reset_vpn_connection(
                context, self.neutron, cleaner, fakes.DB_VPN_GATEWAY_1,
                subnets=[fakes.DB_SUBNET_1],
                route_tables=[fakes.DB_ROUTE_TABLE_1])])

        create_subnet_vpnservice.reset_mock()
        reset_vpn_connection.reset_mock()
        self.add_mock_db_items(self.DB_VPN_GATEWAY_1_DETACHED)
        vpn_gateway_api._start_vpn_in_subnet(
            context, self.neutron, cleaner, copy.deepcopy(fakes.DB_SUBNET_1),
            fakes.DB_VPC_1, fakes.DB_ROUTE_TABLE_1)
        self.assertFalse(create_subnet_vpnservice.called)
        self.assertFalse(reset_vpn_connection.called)

    @mock.patch('ec2api.api.vpn_connection._delete_subnet_vpn')
    @mock.patch('ec2api.api.vpn_gateway._safe_delete_vpnservice')
    def test_stop_vpn_in_subnet(self, delete_vpnservice, delete_subnet_vpn):
        context = base.create_context()
        cleaner = common.OnCrashCleaner()
        mock_manager = mock.Mock()
        mock_manager.attach_mock(delete_vpnservice, 'delete_vpnservice')
        mock_manager.attach_mock(delete_subnet_vpn, 'delete_subnet_vpn')

        self.set_mock_db_items(fakes.DB_VPN_CONNECTION_1,
                               fakes.DB_VPN_CONNECTION_2)
        vpn_gateway_api._stop_vpn_in_subnet(
            context, self.neutron, cleaner, copy.deepcopy(fakes.DB_SUBNET_1))
        mock_manager.has_calls([
            mock.call.delete_subnet_vpn(
                context, self.neutron, cleaner, fakes.DB_SUBNET_1,
                fakes.DB_VPN_CONNECTION_1),
            mock.call.delete_subnet_vpn(
                context, self.neutron, cleaner, fakes.DB_SUBNET_1,
                fakes.DB_VPN_CONNECTION_2),
            mock.call.delete_vpnservice(
                self.neutron, fakes.ID_OS_VPNSERVICE_1,
                fakes.ID_EC2_SUBNET_1)])

        delete_subnet_vpn.reset_mock()
        delete_vpnservice.reset_mock()
        vpn_gateway_api._stop_vpn_in_subnet(
            context, self.neutron, cleaner, self.DB_SUBNET_1_NO_VPN)
        self.assertFalse(delete_subnet_vpn.called)
        self.assertFalse(delete_vpnservice.called)

    def test_create_subnet_vpnservice(self):
        self.neutron.create_vpnservice.side_effect = tools.get_neutron_create(
            'vpnservice', fakes.ID_OS_VPNSERVICE_1)
        context = base.create_context()
        cleaner = common.OnCrashCleaner()

        vpn_gateway_api._create_subnet_vpnservice(
            context, self.neutron, cleaner,
            copy.deepcopy(self.DB_SUBNET_1_NO_VPN), fakes.DB_VPC_1)

        self.neutron.create_vpnservice.assert_called_once_with(
            {'vpnservice': tools.purge_dict(fakes.OS_VPNSERVICE_1,
                                            ('id',))})
        self.db_api.update_item.assert_called_once_with(
            mock.ANY, fakes.DB_SUBNET_1)

        try:
            with common.OnCrashCleaner() as cleaner:
                vpn_gateway_api._create_subnet_vpnservice(
                    context, self.neutron, cleaner,
                    copy.deepcopy(self.DB_SUBNET_1_NO_VPN), fakes.DB_VPC_1)
                raise Exception('fake-exception')
        except Exception as ex:
            if str(ex) != 'fake-exception':
                raise
        self.db_api.update_item.assert_called_with(
            mock.ANY, self.DB_SUBNET_1_NO_VPN)
        self.neutron.delete_vpnservice.assert_called_once_with(
            fakes.ID_OS_VPNSERVICE_1)

    @mock.patch('ec2api.api.vpn_gateway._safe_delete_vpnservice')
    def test_delete_subnet_vpnservice(self, delete_vpnservice):
        context = base.create_context()
        cleaner = common.OnCrashCleaner()

        vpn_gateway_api._delete_subnet_vpnservice(
            context, self.neutron, cleaner, copy.deepcopy(fakes.DB_SUBNET_1))

        self.db_api.update_item.assert_called_once_with(
            mock.ANY, self.DB_SUBNET_1_NO_VPN)

        try:
            with common.OnCrashCleaner() as cleaner:
                vpn_gateway_api._delete_subnet_vpnservice(
                    context, self.neutron, cleaner,
                    copy.deepcopy(fakes.DB_SUBNET_1))
                raise Exception('fake-exception')
        except Exception as ex:
            if str(ex) != 'fake-exception':
                raise
        self.db_api.update_item.assert_called_with(
            mock.ANY, fakes.DB_SUBNET_1)
        self.assertFalse(self.neutron.create_vpnservice.called)

    def test_safe_delete_vpnservice(self):
        vpn_gateway_api._safe_delete_vpnservice(
            self.neutron, fakes.ID_OS_VPNSERVICE_1, fakes.ID_EC2_SUBNET_1)
        self.neutron.delete_vpnservice.assert_called_once_with(
            fakes.ID_OS_VPNSERVICE_1)

        self.neutron.delete_vpnservice.side_effect = (
            neutron_exception.NotFound())
        with fixtures.FakeLogger() as log:
            vpn_gateway_api._safe_delete_vpnservice(
                self.neutron, fakes.ID_OS_VPNSERVICE_1, fakes.ID_EC2_SUBNET_1)
        self.assertEqual(0, len(log.output))

        self.neutron.delete_vpnservice.side_effect = (
            neutron_exception.Conflict())
        with fixtures.FakeLogger() as log:
            vpn_gateway_api._safe_delete_vpnservice(
                self.neutron, fakes.ID_OS_VPNSERVICE_1, fakes.ID_EC2_SUBNET_1)
        self.assertIn(fakes.ID_EC2_SUBNET_1, log.output)
        self.assertIn(fakes.ID_OS_VPNSERVICE_1, log.output)



# Copyright 2014
# The Cloudscaling Group, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy
import logging
import re
from unittest import mock

import fixtures
from lxml import etree

from ec2api.api import ec2utils


def update_dict(dict1, dict2):
    """Get a copy of union of two dicts."""
    res = copy.deepcopy(dict1)
    res.update(dict2)
    return res


def purge_dict(dict1, trash_keys):
    """Get a copy of dict, removed keys."""
    res = copy.deepcopy(dict1)
    for key in trash_keys:
        res.pop(key, None)
    return res


def patch_dict(dict1, dict2, trash_iter):
    """Get a copy of union of two dicts, removed keys."""
    res = update_dict(dict1, dict2)
    res = purge_dict(res, trash_iter)
    return res


def get_db_api_add_item(item_id_dict):
    """Generate db_api.add_item mock function."""

    def db_api_add_item(context, kind, data):
        if isinstance(item_id_dict, dict):
            item_id = item_id_dict[kind]
        else:
            item_id = item_id_dict
        data = update_dict(data, {'id': item_id})
        data.setdefault('os_id')
        data.setdefault('vpc_id')
        return data
    return db_api_add_item


def get_db_api_get_items(*items):
    """Generate db_api.get_items mock function."""

    def db_api_get_items(context, kind):
        return [copy.deepcopy(item)
                for item in items
                if ec2utils.get_ec2_id_kind(item['id']) == kind]
    return db_api_get_items


def get_db_api_get_item_by_id(*items):
    """Generate db_api.get_item_by_id mock function."""

    def db_api_get_item_by_id(context, item_id):
        return next((copy.deepcopy(item)
                     for item in items
                     if item['id'] == item_id),
                    None)
    return db_api_get_item_by_id


def get_db_api_get_items_by_ids(*items):
    """Generate db_api.get_items_by_ids mock function."""

    def db_api_get_items_by_ids(context, item_ids):
        return [copy.deepcopy(item)
                for item in items
                if (item['id'] in item_ids)]
    return db_api_get_items_by_ids


def get_db_api_get_items_ids(*items):
    """Generate db_api.get_items_ids mock function."""

    def db_api_get_items_ids(context, kind, item_ids=None, item_os_ids=None):
        return [(item['id'], item['os_id'])
                for item in items
                if (ec2utils.get_ec2_id_kind(item['id']) == kind and
                    (not item_ids or item['id'] in item_ids) and
                    (not item_os_ids or item['os_id'] in item_os_ids))]
    return db_api_get_items_ids


def get_neutron_create(kind, os_id, addon={}):
    """Generate Neutron create an object mock function."""

    def neutron_create(body):
        body = copy.deepcopy(body)
        body[kind].update(addon)
        body[kind]['id'] = os_id
        return body
    return neutron_create


def get_by_1st_arg_getter(results_dict_by_id, notfound_exception=None):
    """Generate mock function for getter by 1st argurment."""

    def getter(obj_id):
        try:
            return copy.deepcopy(results_dict_by_id[obj_id])
        except KeyError:
            if notfound_exception:
                raise notfound_exception
            else:
                return None
    return getter


def get_by_2nd_arg_getter(results_dict_by_id):
    """Generate mock function for getter by 2nd argurment."""

    def getter(_context, obj_id):
        return copy.deepcopy(results_dict_by_id.get(obj_id))
    return getter


def _safe_copy_parameters(args, kwargs):
    # NOTE(ft): deepcopy fails to copy a complicated mock like
    # neutron client mock or OnCrashCleaner object
    def _safe_copy(obj):
        try:
            return copy.deepcopy(obj)
        except Exception:
            return obj

    args = [_safe_copy(arg)
            for arg in args]
    kwargs = {key: _safe_copy(val)
              for key, val in kwargs.items()}
    return (args, kwargs)


class CopyingMock(mock.MagicMock):
    """Mock class for calls with mutable arguments.

    See https://docs.python.org/3/library/unittest.mock-examples.html#
        coping-with-mutable-arguments
    """

    def __call__(self, *args, **kwargs):
        args, kwargs = _safe_copy_parameters(args, kwargs)
        return super(CopyingMock, self).__call__(*args, **kwargs)


def deepcopy_call_args_saver(destination):
    def side_effect(*args, **kwargs):
        args, kwargs = _safe_copy_parameters(args, kwargs)
        destination.append(mock.call(*args, **kwargs))
    return side_effect


_xml_scheme = re.compile(r'\sxmlns=".*"')


def parse_xml(xml_string):
    xml_string = _xml_scheme.sub('', xml_string.decode("utf-8"))
    xml = etree.fromstring(xml_string)

    def convert_node(node):
        children = list(node)
        if len(children):
            if children[0].tag == 'item':
                val = list(convert_node(child)[1] for child in children)
            else:
                val = dict(convert_node(child) for child in children)
        elif node.tag.endswith('Set'):
            val = []
        else:
            # TODO(ft): do not use private function
            val = (ec2utils._try_convert(node.text)
                   if node.text
                   else node.text)
        return node.tag, val

    return dict([convert_node(xml)])


class KeepingHandler(logging.Handler):

    def __init__(self):
        super(KeepingHandler, self).__init__()
        self._storage = []

    def emit(self, record):
        self._storage.append(record)

    def emit_records_to(self, handlers, record_filter=None):
        for record in self._storage:
            if not record_filter or record_filter.filter(record):
                for handler in handlers:
                    if self != handler:
                        handler.emit(record)


class ScreeningFilter(logging.Filter):

    def __init__(self, name=None):
        self._name = name

    def filter(self, record):
        if self._name is not None and record.name == self._name:
            return False
        return True


class ScreeningLogger(fixtures.Fixture):

    def __init__(self, log_name=None):
        super(ScreeningLogger, self).__init__()
        self.handler = KeepingHandler()
        if log_name:
            self._filter = ScreeningFilter(name=log_name)
        else:
            self._filter = None

    def setUp(self):
        super(ScreeningLogger, self).setUp()
        self.useFixture(fixtures.LogHandler(self.handler))

    def __exit__(self, exc_type, exc_val, exc_tb):
        res = super(ScreeningLogger, self).__exit__(exc_type, exc_val, exc_tb)
        handlers = logging.getLogger().handlers
        if exc_type:
            self.handler.emit_records_to(handlers)
        elif self._filter:
            self.handler.emit_records_to(handlers, self._filter)
        return res


def screen_logs(log_name=None):
    def decorator(func):
        def wrapper(*args, **kwargs):
            with ScreeningLogger(log_name):
                return func(*args, **kwargs)
        return wrapper
    return decorator


screen_unexpected_exception_logs = screen_logs('ec2api.api')
screen_all_logs = screen_logs()



# Copyright 2014
# The Cloudscaling Group, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import copy

import netaddr
from novaclient import exceptions as nova_exception
import six

from ec2api.api import common
from ec2api.api import ec2utils
from ec2api.api import vpn_connection as vpn_connection_api
from ec2api import clients
from ec2api.db import api as db_api
from ec2api import exception
from ec2api.i18n import _


HOST_TARGET = 'host'
VPN_TARGET = 'vpn'


"""Route tables related API implementation
"""


class Validator(common.Validator):

    def igw_or_vgw_id(self, id):
        self.ec2_id(id, ['igw', 'vgw'])


def create_route_table(context, vpc_id):
    vpc = ec2utils.get_db_item(context, vpc_id)
    route_table = _create_route_table(context, vpc)
    return {'routeTable': _format_route_table(context, route_table,
                                              is_main=False)}


def create_route(context, route_table_id, destination_cidr_block,
                 gateway_id=None, instance_id=None,
                 network_interface_id=None,
                 vpc_peering_connection_id=None):
    return _set_route(context, route_table_id, destination_cidr_block,
                      gateway_id, instance_id, network_interface_id,
                      vpc_peering_connection_id, False)


def replace_route(context, route_table_id, destination_cidr_block,
                  gateway_id=None, instance_id=None,
                  network_interface_id=None,
                  vpc_peering_connection_id=None):
    return _set_route(context, route_table_id, destination_cidr_block,
                      gateway_id, instance_id, network_interface_id,
                      vpc_peering_connection_id, True)


def delete_route(context, route_table_id, destination_cidr_block):
    route_table = ec2utils.get_db_item(context, route_table_id)
    for route_index, route in enumerate(route_table['routes']):
        if route['destination_cidr_block'] != destination_cidr_block:
            continue
        if route.get('gateway_id', 0) is None:
            msg = _('cannot remove local route %(destination_cidr_block)s '
                    'in route table %(route_table_id)s')
            msg = msg % {'route_table_id': route_table_id,
                         'destination_cidr_block': destination_cidr_block}
            raise exception.InvalidParameterValue(msg)
        break
    else:
        raise exception.InvalidRouteNotFound(
            route_table_id=route_table_id,
            destination_cidr_block=destination_cidr_block)
    update_target = _get_route_target(route)
    if update_target == VPN_TARGET:
        vpn_gateway = db_api.get_item_by_id(context, route['gateway_id'])
        if (not vpn_gateway or
                vpn_gateway['vpc_id'] != route_table['vpc_id']):
            update_target = None
    rollback_route_table_state = copy.deepcopy(route_table)
    del route_table['routes'][route_index]
    with common.OnCrashCleaner() as cleaner:
        db_api.update_item(context, route_table)
        cleaner.addCleanup(db_api.update_item, context,
                           rollback_route_table_state)

        if update_target:
            _update_routes_in_associated_subnets(
                context, cleaner, route_table, update_target=update_target)

    return True


def enable_vgw_route_propagation(context, route_table_id, gateway_id):
    route_table = ec2utils.get_db_item(context, route_table_id)
    # NOTE(ft): AWS returns GatewayNotAttached for all invalid cases of
    # gateway_id value
    vpn_gateway = ec2utils.get_db_item(context, gateway_id)
    if vpn_gateway['vpc_id'] != route_table['vpc_id']:
        raise exception.GatewayNotAttached(gw_id=vpn_gateway['id'],
                                           vpc_id=route_table['vpc_id'])
    if vpn_gateway['id'] in route_table.setdefault('propagating_gateways', []):
        return True
    with common.OnCrashCleaner() as cleaner:
        _append_propagation_to_route_table_item(context, route_table,
                                                vpn_gateway['id'])
        cleaner.addCleanup(_remove_propagation_from_route_table_item,
                           context, route_table, vpn_gateway['id'])

        _update_routes_in_associated_subnets(context, cleaner, route_table,
                                             update_target=VPN_TARGET)
    return True


def disable_vgw_route_propagation(context, route_table_id, gateway_id):
    route_table = ec2utils.get_db_item(context, route_table_id)
    if gateway_id not in route_table.get('propagating_gateways', []):
        return True
    vpn_gateway = db_api.get_item_by_id(context, gateway_id)

    with common.OnCrashCleaner() as cleaner:
        _remove_propagation_from_route_table_item(context, route_table,
                                                  gateway_id)
        cleaner.addCleanup(_append_propagation_to_route_table_item,
                           context, route_table, gateway_id)

        if vpn_gateway and vpn_gateway['vpc_id'] == route_table['vpc_id']:
            _update_routes_in_associated_subnets(context, cleaner, route_table,
                                                 update_target=VPN_TARGET)
    return True


def associate_route_table(context, route_table_id, subnet_id):
    route_table = ec2utils.get_db_item(context, route_table_id)
    subnet = ec2utils.get_db_item(context, subnet_id)
    if route_table['vpc_id'] != subnet['vpc_id']:
        msg = _('Route table %(rtb_id)s and subnet %(subnet_id)s belong to '
                'different networks')
        msg = msg % {'rtb_id': route_table_id,
                     'subnet_id': subnet_id}
        raise exception.InvalidParameterValue(msg)
    if 'route_table_id' in subnet:
        msg = _('The specified association for route table %(rtb_id)s '
                'conflicts with an existing association')
        msg = msg % {'rtb_id': route_table_id}
        raise exception.ResourceAlreadyAssociated(msg)

    with common.OnCrashCleaner() as cleaner:
        _associate_subnet_item(context, subnet, route_table['id'])
        cleaner.addCleanup(_disassociate_subnet_item, context, subnet)

        _update_subnet_routes(context, cleaner, subnet, route_table)

    return {'associationId': ec2utils.change_ec2_id_kind(subnet['id'],
                                                         'rtbassoc')}


def replace_route_table_association(context, association_id, route_table_id):
    route_table = ec2utils.get_db_item(context, route_table_id)
    if route_table['vpc_id'] == ec2utils.change_ec2_id_kind(association_id,
                                                            'vpc'):
        vpc = db_api.get_item_by_id(
            context, ec2utils.change_ec2_id_kind(association_id, 'vpc'))
        if vpc is None:
            raise exception.InvalidAssociationIDNotFound(id=association_id)

        rollback_route_table_id = vpc['route_table_id']
        with common.OnCrashCleaner() as cleaner:
            _associate_vpc_item(context, vpc, route_table['id'])
            cleaner.addCleanup(_associate_vpc_item, context, vpc,
                               rollback_route_table_id)

            _update_routes_in_associated_subnets(
                context, cleaner, route_table, default_associations_only=True)
    else:
        subnet = db_api.get_item_by_id(
            context, ec2utils.change_ec2_id_kind(association_id, 'subnet'))
        if subnet is None or 'route_table_id' not in subnet:
            raise exception.InvalidAssociationIDNotFound(id=association_id)
        if subnet['vpc_id'] != route_table['vpc_id']:
            msg = _('Route table association %(rtbassoc_id)s and route table '
                    '%(rtb_id)s belong to different networks')
            msg = msg % {'rtbassoc_id': association_id,
                         'rtb_id': route_table_id}
            raise exception.InvalidParameterValue(msg)

        rollback_route_table_id = subnet['route_table_id']
        with common.OnCrashCleaner() as cleaner:
            _associate_subnet_item(context, subnet, route_table['id'])
            cleaner.addCleanup(_associate_subnet_item, context, subnet,
                               rollback_route_table_id)

            _update_subnet_routes(context, cleaner, subnet, route_table)

    return {'newAssociationId': association_id}


def disassociate_route_table(context, association_id):
    subnet = db_api.get_item_by_id(
        context, ec2utils.change_ec2_id_kind(association_id, 'subnet'))
    if not subnet:
        vpc = db_api.get_item_by_id(
            context, ec2utils.change_ec2_id_kind(association_id, 'vpc'))
        if vpc is None:
            raise exception.InvalidAssociationIDNotFound(id=association_id)
        msg = _('Cannot disassociate the main route table association '
                '%(rtbassoc_id)s') % {'rtbassoc_id': association_id}
        raise exception.InvalidParameterValue(msg)
    if 'route_table_id' not in subnet:
        raise exception.InvalidAssociationIDNotFound(id=association_id)

    rollback_route_table_id = subnet['route_table_id']
    vpc = db_api.get_item_by_id(context, subnet['vpc_id'])
    main_route_table = db_api.get_item_by_id(context, vpc['route_table_id'])
    with common.OnCrashCleaner() as cleaner:
        _disassociate_subnet_item(context, subnet)
        cleaner.addCleanup(_associate_subnet_item, context, subnet,
                           rollback_route_table_id)

        _update_subnet_routes(context, cleaner, subnet, main_route_table)

    return True


def delete_route_table(context, route_table_id):
    route_table = ec2utils.get_db_item(context, route_table_id)
    vpc = db_api.get_item_by_id(context, route_table['vpc_id'])
    _delete_route_table(context, route_table['id'], vpc)
    return True


class RouteTableDescriber(common.TaggableItemsDescriber,
                          common.NonOpenstackItemsDescriber):

    KIND = 'rtb'
    FILTER_MAP = {'association.route-table-association-id': (
                        ['associationSet', 'routeTableAssociationId']),
                  'association.route-table-id': ['associationSet',
                                                 'routeTableId'],
                  'association.subnet-id': ['associationSet', 'subnetId'],
                  'association.main': ['associationSet', 'main'],
                  'route-table-id': 'routeTableId',
                  'route.destination-cidr-block': ['routeSet',
                                                   'destinationCidrBlock'],
                  'route.gateway-id': ['routeSet', 'gatewayId'],
                  'route.instance-id': ['routeSet', 'instanceId'],
                  'route.origin': ['routeSet', 'origin'],
                  'route.state': ['routeSet', 'state'],
                  'vpc-id': 'vpcId'}

    def format(self, route_table):
        return _format_route_table(
            self.context, route_table,
            associated_subnet_ids=self.associations[route_table['id']],
            is_main=(self.vpcs[route_table['vpc_id']]['route_table_id'] ==
                     route_table['id']),
            gateways=self.gateways,
            network_interfaces=self.network_interfaces,
            vpn_connections_by_gateway_id=self.vpn_connections_by_gateway_id)

    def get_db_items(self):
        associations = collections.defaultdict(list)
        for subnet in db_api.get_items(self.context, 'subnet'):
            if 'route_table_id' in subnet:
                associations[subnet['route_table_id']].append(subnet['id'])
        self.associations = associations
        vpcs = db_api.get_items(self.context, 'vpc')
        self.vpcs = {vpc['id']: vpc for vpc in vpcs}
        gateways = (db_api.get_items(self.context, 'igw') +
                    db_api.get_items(self.context, 'vgw'))
        self.gateways = {gw['id']: gw for gw in gateways}
        # TODO(ft): scan route tables to get only used instances and
        # network interfaces to reduce DB and Nova throughput
        network_interfaces = db_api.get_items(self.context, 'eni')
        self.network_interfaces = {eni['id']: eni
                                   for eni in network_interfaces}
        vpn_connections = db_api.get_items(self.context, 'vpn')
        vpns_by_gateway_id = {}
        for vpn in vpn_connections:
            vpns = vpns_by_gateway_id.setdefault(vpn['vpn_gateway_id'], [])
            vpns.append(vpn)
        self.vpn_connections_by_gateway_id = vpns_by_gateway_id
        return super(RouteTableDescriber, self).get_db_items()


def describe_route_tables(context, route_table_id=None, filter=None):
    ec2utils.check_and_create_default_vpc(context)
    formatted_route_tables = RouteTableDescriber().describe(
            context, ids=route_table_id, filter=filter)
    return {'routeTableSet': formatted_route_tables}


def _create_route_table(context, vpc):
    route_table = {'vpc_id': vpc['id'],
                   'routes': [{'destination_cidr_block': vpc['cidr_block'],
                               'gateway_id': None}]}
    route_table = db_api.add_item(context, 'rtb', route_table)
    return route_table


def _delete_route_table(context, route_table_id, vpc=None, cleaner=None):
    def get_associated_subnets():
        return [s for s in db_api.get_items(context, 'subnet')
                if s.get('route_table_id') == route_table_id]

    if (vpc and route_table_id == vpc['route_table_id'] or
            len(get_associated_subnets()) > 0):
        msg = _("The routeTable '%(rtb_id)s' has dependencies and cannot "
                "be deleted.") % {'rtb_id': route_table_id}
        raise exception.DependencyViolation(msg)
    if cleaner:
        route_table = db_api.get_item_by_id(context, route_table_id)
    db_api.delete_item(context, route_table_id)
    if cleaner and route_table:
        cleaner.addCleanup(db_api.restore_item, context, 'rtb', route_table)


def _set_route(context, route_table_id, destination_cidr_block,
               gateway_id, instance_id, network_interface_id,
               vpc_peering_connection_id, do_replace):
    route_table = ec2utils.get_db_item(context, route_table_id)
    vpc = db_api.get_item_by_id(context, route_table['vpc_id'])
    vpc_ipnet = netaddr.IPNetwork(vpc['cidr_block'])
    route_ipnet = netaddr.IPNetwork(destination_cidr_block)
    if route_ipnet in vpc_ipnet:
        msg = _('Cannot create a more specific route for '
                '%(destination_cidr_block)s than local route '
                '%(vpc_cidr_block)s in route table %(rtb_id)s')
        msg = msg % {'rtb_id': route_table_id,
                     'destination_cidr_block': destination_cidr_block,
                     'vpc_cidr_block': vpc['cidr_block']}
        raise exception.InvalidParameterValue(msg)

    obj_param_count = len([p for p in (gateway_id, network_interface_id,
                                       instance_id, vpc_peering_connection_id)
                           if p is not None])
    if obj_param_count != 1:
        msg = _('The request must contain exactly one of gatewayId, '
                'networkInterfaceId, vpcPeeringConnectionId or instanceId')
        if obj_param_count == 0:
            raise exception.MissingParameter(msg)
        else:
            raise exception.InvalidParameterCombination(msg)

    rollabck_route_table_state = copy.deepcopy(route_table)
    if do_replace:
        route_index, old_route = next(
            ((i, r) for i, r in enumerate(route_table['routes'])
             if r['destination_cidr_block'] == destination_cidr_block),
            (None, None))
        if route_index is None:
            msg = _("There is no route defined for "
                    "'%(destination_cidr_block)s' in the route table. "
                    "Use CreateRoute instead.")
            msg = msg % {'destination_cidr_block': destination_cidr_block}
            raise exception.InvalidParameterValue(msg)
        else:
            del route_table['routes'][route_index]

    if gateway_id:
        gateway = ec2utils.get_db_item(context, gateway_id)
        if gateway.get('vpc_id') != route_table['vpc_id']:
            if ec2utils.get_ec2_id_kind(gateway_id) == 'vgw':
                raise exception.InvalidGatewayIDNotFound(id=gateway['id'])
            else:  # igw
                raise exception.InvalidParameterValue(
                    _('Route table %(rtb_id)s and network gateway %(igw_id)s '
                      'belong to different networks') %
                    {'rtb_id': route_table_id,
                     'igw_id': gateway_id})
        route = {'gateway_id': gateway['id']}
    elif network_interface_id:
        network_interface = ec2utils.get_db_item(context, network_interface_id)
        if network_interface['vpc_id'] != route_table['vpc_id']:
            msg = _('Route table %(rtb_id)s and interface %(eni_id)s '
                    'belong to different networks')
            msg = msg % {'rtb_id': route_table_id,
                         'eni_id': network_interface_id}
            raise exception.InvalidParameterValue(msg)
        route = {'network_interface_id': network_interface['id']}
    elif instance_id:
        # TODO(ft): implement search in DB layer
        network_interfaces = [eni for eni in db_api.get_items(context, 'eni')
                              if eni.get('instance_id') == instance_id]
        if len(network_interfaces) == 0:
            msg = _("Invalid value '%(i_id)s' for instance ID. "
                    "Instance is not in a VPC.")
            msg = msg % {'i_id': instance_id}
            raise exception.InvalidParameterValue(msg)
        elif len(network_interfaces) > 1:
            raise exception.InvalidInstanceId(instance_id=instance_id)
        network_interface = network_interfaces[0]
        if network_interface['vpc_id'] != route_table['vpc_id']:
            msg = _('Route table %(rtb_id)s and interface %(eni_id)s '
                    'belong to different networks')
            msg = msg % {'rtb_id': route_table_id,
                         'eni_id': network_interface['id']}
            raise exception.InvalidParameterValue(msg)
        route = {'network_interface_id': network_interface['id']}
    else:
        raise exception.InvalidRequest('Parameter VpcPeeringConnectionId is '
                                       'not supported by this implementation')
    route['destination_cidr_block'] = destination_cidr_block
    update_target = _get_route_target(route)

    if do_replace:
        idempotent_call = False
        old_target = _get_route_target(old_route)
        if old_target != update_target:
            update_target = None
    else:
        old_route = next((r for r in route_table['routes']
                          if r['destination_cidr_block'] ==
                          destination_cidr_block), None)
        idempotent_call = old_route == route
        if old_route and not idempotent_call:
            raise exception.RouteAlreadyExists(
                destination_cidr_block=destination_cidr_block)

    if not idempotent_call:
        route_table['routes'].append(route)

    with common.OnCrashCleaner() as cleaner:
        db_api.update_item(context, route_table)
        cleaner.addCleanup(db_api.update_item, context,
                           rollabck_route_table_state)
        _update_routes_in_associated_subnets(context, cleaner, route_table,
                                             update_target=update_target)

    return True


def _format_route_table(context, route_table, is_main=False,
                        associated_subnet_ids=[],
                        gateways={},
                        network_interfaces={},
                        vpn_connections_by_gateway_id={}):
    vpc_id = route_table['vpc_id']
    ec2_route_table = {
        'routeTableId': route_table['id'],
        'vpcId': vpc_id,
        'routeSet': [],
        'propagatingVgwSet': [
            {'gatewayId': vgw_id}
            for vgw_id in route_table.get('propagating_gateways', [])],
        # NOTE(ft): AWS returns empty tag set for a route table
        # if no tag exists
        'tagSet': [],
    }
    # TODO(ft): refactor to get Nova instances outside of this function
    nova = clients.nova(context)
    for route in route_table['routes']:
        origin = ('CreateRouteTable'
                  if route.get('gateway_id', 0) is None else
                  'CreateRoute')
        ec2_route = {'destinationCidrBlock': route['destination_cidr_block'],
                     'origin': origin}
        if 'gateway_id' in route:
            gateway_id = route['gateway_id']
            if gateway_id is None:
                state = 'active'
                ec2_gateway_id = 'local'
            else:
                gateway = gateways.get(gateway_id)
                state = ('active'
                         if gateway and gateway.get('vpc_id') == vpc_id else
                         'blackhole')
                ec2_gateway_id = gateway_id
            ec2_route.update({'gatewayId': ec2_gateway_id,
                              'state': state})
        else:
            network_interface_id = route['network_interface_id']
            network_interface = network_interfaces.get(network_interface_id)
            instance_id = (network_interface.get('instance_id')
                           if network_interface else
                           None)
            state = 'blackhole'
            if instance_id:
                instance = db_api.get_item_by_id(context, instance_id)
                if instance:
                    try:
                        os_instance = nova.servers.get(instance['os_id'])
                        if os_instance and os_instance.status == 'ACTIVE':
                            state = 'active'
                    except nova_exception.NotFound:
                        pass
                ec2_route.update({'instanceId': instance_id,
                                  'instanceOwnerId': context.project_id})
            ec2_route.update({'networkInterfaceId': network_interface_id,
                              'state': state})
        ec2_route_table['routeSet'].append(ec2_route)

    for vgw_id in route_table.get('propagating_gateways', []):
        vgw = gateways.get(vgw_id)
        if vgw and vgw_id in vpn_connections_by_gateway_id:
            cidrs = set()
            vpn_connections = vpn_connections_by_gateway_id[vgw_id]
            for vpn_connection in vpn_connections:
                cidrs.update(vpn_connection['cidrs'])
            state = 'active' if vgw['vpc_id'] == vpc_id else 'blackhole'
            for cidr in cidrs:
                ec2_route = {'gatewayId': vgw_id,
                             'destinationCidrBlock': cidr,
                             'state': state,
                             'origin': 'EnableVgwRoutePropagation'}
                ec2_route_table['routeSet'].append(ec2_route)

    associations = []
    if is_main:
        associations.append({
            'routeTableAssociationId': ec2utils.change_ec2_id_kind(vpc_id,
                                                                   'rtbassoc'),
            'routeTableId': route_table['id'],
            'main': True})
    for subnet_id in associated_subnet_ids:
        associations.append({
            'routeTableAssociationId': ec2utils.change_ec2_id_kind(subnet_id,
                                                                   'rtbassoc'),
            'routeTableId': route_table['id'],
            'subnetId': subnet_id,
            'main': False})
    if associations:
        ec2_route_table['associationSet'] = associations

    return ec2_route_table


def _update_routes_in_associated_subnets(context, cleaner, route_table,
                                         default_associations_only=None,
                                         update_target=None):
    if default_associations_only:
        appropriate_rtb_ids = (None,)
    else:
        vpc = db_api.get_item_by_id(context, route_table['vpc_id'])
        if vpc['route_table_id'] == route_table['id']:
            appropriate_rtb_ids = (route_table['id'], None)
        else:
            appropriate_rtb_ids = (route_table['id'],)
    neutron = clients.neutron(context)
    subnets = [subnet for subnet in db_api.get_items(context, 'subnet')
               if (subnet['vpc_id'] == route_table['vpc_id'] and
                   subnet.get('route_table_id') in appropriate_rtb_ids)]
    # NOTE(ft): we need to update host routes for both host and vpn target
    # because vpn-related routes are present in host routes as well
    _update_host_routes(context, neutron, cleaner, route_table, subnets)
    if not update_target or update_target == VPN_TARGET:
        vpn_connection_api._update_vpn_routes(context, neutron, cleaner,
                                              route_table, subnets)


def _update_subnet_routes(context, cleaner, subnet, route_table):
    neutron = clients.neutron(context)
    _update_host_routes(context, neutron, cleaner, route_table, [subnet])
    vpn_connection_api._update_vpn_routes(context, neutron, cleaner,
                                          route_table, [subnet])


def _update_host_routes(context, neutron, cleaner, route_table, subnets):
    destinations = _get_active_route_destinations(context, route_table)
    for subnet in subnets:
        # TODO(ft): do list subnet w/ filters instead of show one by one
        os_subnet = neutron.show_subnet(subnet['os_id'])['subnet']
        host_routes, gateway_ip = _get_subnet_host_routes_and_gateway_ip(
            context, route_table, os_subnet['cidr'], destinations)
        neutron.update_subnet(subnet['os_id'],
                              {'subnet': {'host_routes': host_routes,
                                          'gateway_ip': gateway_ip}})
        cleaner.addCleanup(
            neutron.update_subnet, subnet['os_id'],
            {'subnet': {'host_routes': os_subnet['host_routes'],
                        'gateway_ip': os_subnet['gateway_ip']}})


def _get_active_route_destinations(context, route_table):
    vpn_connections = {vpn['vpn_gateway_id']: vpn
                       for vpn in db_api.get_items(context, 'vpn')}
    dst_ids = [route[id_key]
               for route in route_table['routes']
               for id_key in ('gateway_id', 'network_interface_id')
               if route.get(id_key) is not None]
    dst_ids.extend(route_table.get('propagating_gateways', []))
    destinations = {item['id']: item
                    for item in db_api.get_items_by_ids(context, dst_ids)
                    if (item['vpc_id'] == route_table['vpc_id'] and
                        (ec2utils.get_ec2_id_kind(item['id']) != 'vgw' or
                         item['id'] in vpn_connections))}
    for vpn in six.itervalues(vpn_connections):
        if vpn['vpn_gateway_id'] in destinations:
            destinations[vpn['vpn_gateway_id']]['vpn_connection'] = vpn
    return destinations


def _get_subnet_host_routes_and_gateway_ip(context, route_table, cidr_block,
                                           destinations=None):
    if not destinations:
        destinations = _get_active_route_destinations(context, route_table)
    gateway_ip = str(netaddr.IPAddress(
        netaddr.IPNetwork(cidr_block).first + 1))

    def get_nexthop(route):
        if 'gateway_id' in route:
            gateway_id = route['gateway_id']
            if gateway_id and gateway_id not in destinations:
                return '127.0.0.1'
            return gateway_ip
        network_interface = destinations.get(route['network_interface_id'])
        if not network_interface:
            return '127.0.0.1'
        return network_interface['private_ip_address']

    host_routes = []
    subnet_gateway_is_used = False
    for route in route_table['routes']:
        nexthop = get_nexthop(route)
        cidr = route['destination_cidr_block']
        if cidr == '0.0.0.0/0':
            if nexthop == '127.0.0.1':
                continue
            elif nexthop == gateway_ip:
                subnet_gateway_is_used = True
        host_routes.append({'destination': cidr,
                            'nexthop': nexthop})
    host_routes.extend(
        {'destination': cidr,
         'nexthop': gateway_ip}
        for vgw_id in route_table.get('propagating_gateways', [])
        for cidr in (destinations.get(vgw_id, {}).get('vpn_connection', {}).
                     get('cidrs', [])))

    if not subnet_gateway_is_used:
        # NOTE(andrey-mp): add route to metadata server
        host_routes.append(
            {'destination': '169.254.169.254/32',
             'nexthop': gateway_ip})
        # NOTE(ft): gateway_ip is set to None to allow correct handling
        # of 0.0.0.0/0 route by Neutron.
        gateway_ip = None
    return host_routes, gateway_ip


def _get_route_target(route):
    if ec2utils.get_ec2_id_kind(route.get('gateway_id') or '') == 'vgw':
        return VPN_TARGET
    else:
        return HOST_TARGET


def _associate_subnet_item(context, subnet, route_table_id):
    subnet['route_table_id'] = route_table_id
    db_api.update_item(context, subnet)


def _disassociate_subnet_item(context, subnet):
    subnet.pop('route_table_id')
    db_api.update_item(context, subnet)


def _associate_vpc_item(context, vpc, route_table_id):
    vpc['route_table_id'] = route_table_id
    db_api.update_item(context, vpc)


def _append_propagation_to_route_table_item(context, route_table, gateway_id):
    vgws = route_table.setdefault('propagating_gateways', [])
    vgws.append(gateway_id)
    db_api.update_item(context, route_table)


def _remove_propagation_from_route_table_item(context, route_table,
                                              gateway_id):
    vgws = route_table['propagating_gateways']
    vgws.remove(gateway_id)
    if not vgws:
        del route_table['propagating_gateways']
    db_api.update_item(context, route_table)



# Copyright 2014
# The Cloudscaling Group, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import collections
import fnmatch
import inspect
import operator

from oslo_config import cfg
from oslo_log import log as logging
import six

from ec2api.api import ec2utils
from ec2api.api import validator
from ec2api.db import api as db_api
from ec2api import exception
from ec2api.i18n import _


ec2_opts = [
    cfg.BoolOpt('disable_ec2_classic',
                help='True if server does not support EC2 Classic mode '
                     'in favor of default VPC'),
]

CONF = cfg.CONF
CONF.register_opts(ec2_opts)
LOG = logging.getLogger(__name__)


class OnCrashCleaner(object):

    def __init__(self):
        self._cleanups = []
        self._suppress_exception = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            return
        self._run_cleanups(self._cleanups)
        return self._suppress_exception

    def addCleanup(self, function, *args, **kwargs):
        self._cleanups.append((function, args, kwargs))

    def approveChanges(self):
        del self._cleanups[:]
        self._suppress_exception = True

    def _run_cleanups(self, cleanups):
        for function, args, kwargs in reversed(cleanups):
            try:
                function(*args, **kwargs)
            except Exception:
                if inspect.ismethod(function):
                    if six.PY2:
                        cmodule = function.im_class.__module__
                        cname = function.im_class.__name__
                    else:
                        cmodule = function.__self__.__class__.__module__
                        cname = function.__self__.__class__.__name__
                    name = '%s.%s.%s' % (cmodule, cname, function.__name__)
                elif inspect.isfunction(function):
                    name = '%s.%s' % (function.__module__, function.__name__)
                else:
                    name = '%s.%s' % (function.__class__.__module__,
                                      function.__class__.__name__)
                formatted_args = ''
                args_string = ', '.join([repr(arg) for arg in args])
                kwargs_string = ', '.join([
                    '%s=%r' % (key, value) for key, value in kwargs.items()
                ])
                if args_string:
                    formatted_args = args_string
                if kwargs_string:
                    if formatted_args:
                        formatted_args += ', '
                    formatted_args += kwargs_string
                LOG.warning(
                    'Error cleaning up %(name)s(%(args)s)' %
                    {'name': name, 'args': formatted_args},
                    exc_info=True)
                pass


class Validator(object):

    def __init__(self, param_name="", action="", params=[]):
        self.param_name = param_name
        self.action = action
        self.params = params

    def multi(self, items, validation_func):
        validator.validate_list(items, self.param_name)
        for item in items:
            validation_func(item)

    def dummy(self, value):
        pass

    def bool(self, value):
        validator.validate_bool(value, self.param_name)

    def int(self, value):
        validator.validate_int(value, self.param_name)

    def str(self, value):
        validator.validate_str(value, self.param_name)

    def strs(self, values):
        self.multi(values, self.str)

    def str64(self, value):
        validator.validate_str(value, self.param_name, 64)

    def str255(self, value):
        validator.validate_str(value, self.param_name, 255)

    def str255s(self, values):
        self.multi(values, self.str255)

    def ip(self, ip):
        validator.validate_ipv4(ip, self.param_name)

    def ips(self, ips):
        self.multi(ips, self.ip)

    def cidr(self, cidr):
        validator.validate_cidr(cidr, self.param_name)

    def subnet_cidr(self, cidr):
        validator.validate_subnet_cidr(cidr)

    def vpc_cidr(self, cidr):
        validator.validate_vpc_cidr(cidr)

    def filter(self, filter):
        validator.validate_filter(filter)

    def key_value_dict_list(self, dict_list):
        validator.validate_key_value_dict_list(dict_list, self.param_name)

    def ec2_id(self, id, prefices=None):
        validator.validate_ec2_id(id, self.param_name, prefices)

    def ec2_ids(self, ids):
        self.multi(ids, self.ec2_id)

    def i_id(self, id):
        self.ec2_id(id, ['i'])

    def i_ids(self, ids):
        self.multi(ids, self.i_id)

    def ami_id(self, id):
        self.ec2_id(id, ['ami'])

    def aki_id(self, id):
        self.ec2_id(id, ['aki'])

    def ari_id(self, id):
        self.ec2_id(id, ['ari'])

    def amiariaki_id(self, id):
        self.ec2_id(id, ['ami', 'ari', 'aki'])

    def amiariaki_ids(self, ids):
        self.multi(ids, self.amiariaki_id)

    def sg_id(self, id):
        self.ec2_id(id, ['sg'])

    def sg_ids(self, ids):
        self.multi(ids, self.sg_id)

    def subnet_id(self, id):
        self.ec2_id(id, ['subnet'])

    def subnet_ids(self, ids):
        self.multi(ids, self.subnet_id)

    def igw_id(self, id):
        self.ec2_id(id, ['igw'])

    def igw_ids(self, ids):
        self.multi(ids, self.igw_id)

    def rtb_id(self, id):
        self.ec2_id(id, ['rtb'])

    def rtb_ids(self, ids):
        self.multi(ids, self.rtb_id)

    def eni_id(self, id):
        self.ec2_id(id, ['eni'])

    def eni_ids(self, ids):
        self.multi(ids, self.eni_id)

    def vpc_id(self, id):
        self.ec2_id(id, ['vpc'])

    def vpc_ids(self, ids):
        self.multi(ids, self.vpc_id)

    def eipalloc_id(self, id):
        self.ec2_id(id, ['eipalloc'])

    def eipalloc_ids(self, ids):
        self.multi(ids, self.eipalloc_id)

    def eipassoc_id(self, id):
        self.ec2_id(id, ['eipassoc'])

    def rtbassoc_id(self, id):
        self.ec2_id(id, ['rtbassoc'])

    def eni_attach_id(self, id):
        self.ec2_id(id, ['eni-attach'])

    def snap_id(self, id):
        self.ec2_id(id, ['snap'])

    def snap_ids(self, ids):
        self.multi(ids, self.snap_id)

    def vol_id(self, id):
        self.ec2_id(id, ['vol'])

    def vol_ids(self, ids):
        self.multi(ids, self.vol_id)

    def dopt_id(self, id):
        self.ec2_id(id, ['dopt'])

    def dopt_ids(self, ids):
        self.multi(ids, self.dopt_id)

    def vgw_id(self, id):
        self.ec2_id(id, ['vgw'])

    def vgw_ids(self, ids):
        self.multi(ids, self.vgw_id)

    def cgw_id(self, id):
        self.ec2_id(id, ['cgw'])

    def cgw_ids(self, ids):
        self.multi(ids, self.cgw_id)

    def vpn_id(self, id):
        self.ec2_id(id, ['vpn'])

    def vpn_ids(self, ids):
        self.multi(ids, self.vpn_id)

    def security_group_str(self, value):
        validator.validate_security_group_str(value, self.param_name,
                                              self.params.get('vpc_id'))

    def security_group_strs(self, values):
        self.multi(values, self.security_group_str)

    def vpn_connection_type(self, value):
        validator.validate_vpn_connection_type(value)


VPC_KINDS = ['vpc', 'igw', 'subnet', 'eni', 'dopt', 'eipalloc', 'rtb',
             'vgw', 'cgw', 'vpn']


class UniversalDescriber(object):
    """Abstract Describer class for various Describe implementations."""

    KIND = ''
    SORT_KEY = ''
    FILTER_MAP = {}

    def format(self, item=None, os_item=None):
        pass

    def post_format(self, formatted_item, item):
        pass

    def get_db_items(self):
        return ec2utils.get_db_items(self.context, self.KIND, self.ids)

    def get_os_items(self):
        return []

    def auto_update_db(self, item, os_item):
        if item is None and self.KIND not in VPC_KINDS:
            item = ec2utils.auto_create_db_item(self.context, self.KIND,
                                                self.get_id(os_item))
            LOG.info(
                'Item %(item)s was updated to %(os_item)s.',
                {'item': str(item), 'os_item': str(os_item)})
        return item

    def get_id(self, os_item):
        return os_item['id'] if isinstance(os_item, dict) else os_item.id

    def get_name(self, os_item):
        return os_item['name']

    def delete_obsolete_item(self, item):
        LOG.info('Deleting obsolete item %(item)s', {'item': str(item)})
        db_api.delete_item(self.context, item['id'])

    def is_filtering_value_found(self, filter_value, value):
        if fnmatch.fnmatch(str(value), str(filter_value)):
            return True

    def filtered_out(self, item, filters):
        if filters is None:
            return False
        for filter in filters:
            filter_name = self.FILTER_MAP.get(filter['name'])
            if filter_name is None:
                raise exception.InvalidParameterValue(
                    value=filter['name'], parameter='filter',
                    reason='invalid filter')
            values = self.get_values_by_filter(filter_name, item)
            if not values:
                return True
            filter_values = filter['value']
            for filter_value in filter_values:
                if any(self.is_filtering_value_found(filter_value, value)
                       for value in values):
                    break
            else:
                return True
        return False

    def get_values_by_filter(self, filter_name, item):
        if isinstance(filter_name, list):
            values = []
            value_set = item.get(filter_name[0], [])
            for value in value_set:
                vals = self.get_values_by_filter(filter_name[1], value)
                if vals:
                    values += vals
        else:
            if isinstance(filter_name, tuple):
                value = item.get(filter_name[0], {}).get(filter_name[1])
            else:
                value = item.get(filter_name)
            values = [value] if value is not None else []
        return values

    def get_paged(self, formatted_items, max_results, next_token):
        self.next_token = None
        if not max_results and not next_token:
            return formatted_items

        if max_results and max_results > 1000:
            max_results = 1000
        formatted_items = sorted(formatted_items,
                                 key=operator.itemgetter(self.SORT_KEY))

        next_item = 0
        if next_token:
            next_item = int(base64.b64decode(next_token).decode())
        if next_item:
            formatted_items = formatted_items[next_item:]
        if max_results and max_results < len(formatted_items):
            self.next_token = base64.b64encode(
                str(next_item + max_results).encode())
            formatted_items = formatted_items[:max_results]

        return formatted_items

    def is_selected_item(self, context, os_item_name, item):
        return (os_item_name in self.names or
                (item and item['id'] in self.ids))

    def handle_unpaired_item(self, item):
        self.delete_obsolete_item(item)

    def describe(self, context, ids=None, names=None, filter=None,
                 max_results=None, next_token=None):
        if max_results and max_results < 5:
            msg = (_('Value ( %s ) for parameter maxResults is invalid. '
                     'Expecting a value greater than 5.') % max_results)
            raise exception.InvalidParameterValue(msg)

        self.context = context
        self.selective_describe = ids is not None or names is not None
        self.ids = set(ids or [])
        self.names = set(names or [])
        self.items = self.get_db_items()
        self.os_items = self.get_os_items()
        formatted_items = []

        self.items_dict = {i['os_id']: i for i in (self.items or [])}
        paired_items_ids = set()
        for os_item in self.os_items:
            os_item_name = self.get_name(os_item)
            os_item_id = self.get_id(os_item)
            item = self.items_dict.get(os_item_id, None)
            if item:
                paired_items_ids.add(item['id'])
            # NOTE(Alex): Filter out items not requested in names or ids
            if (self.selective_describe and
                    not self.is_selected_item(context, os_item_name, item)):
                continue
            # NOTE(Alex): Autoupdate DB for autoupdatable items
            item = self.auto_update_db(item, os_item)
            # NOTE(andrey-mp): save item id again
            # (if item has created by auto update)
            if item:
                paired_items_ids.add(item['id'])
            formatted_item = self.format(item, os_item)
            self.post_format(formatted_item, item)
            if os_item_name in self.names:
                self.names.remove(os_item_name)
            if item and item['id'] in self.ids:
                self.ids.remove(item['id'])
            if (formatted_item and
                    not self.filtered_out(formatted_item, filter)):
                formatted_items.append(formatted_item)
        # NOTE(Alex): delete obsolete items
        for item in self.items:
            if item['id'] in paired_items_ids:
                continue
            formatted_item = self.handle_unpaired_item(item)
            if formatted_item:
                if not self.filtered_out(formatted_item, filter):
                    formatted_items.append(formatted_item)
                if item['id'] in self.ids:
                    self.ids.remove(item['id'])
        # NOTE(Alex): some requested items are not found
        if self.ids or self.names:
            params = {'id': next(iter(self.ids or self.names))}
            raise ec2utils.NOT_FOUND_EXCEPTION_MAP[self.KIND](**params)

        return self.get_paged(formatted_items, max_results, next_token)


class TaggableItemsDescriber(UniversalDescriber):

    tags = None

    def __init__(self):
        super(TaggableItemsDescriber, self).__init__()
        self.FILTER_MAP['tag-key'] = ['tagSet', 'key']
        self.FILTER_MAP['tag-value'] = ['tagSet', 'value']
        self.FILTER_MAP['tag'] = 'tagSet'

    def get_tags(self):
        return db_api.get_tags(self.context, (self.KIND,), self.ids)

    def post_format(self, formatted_item, item):
        if not item or not formatted_item:
            return

        if self.tags is None:
            tags = collections.defaultdict(list)
            for tag in self.get_tags():
                tags[tag['item_id']].append(tag)
            self.tags = tags

        formatted_tags = []
        for tag in self.tags[item['id']]:
            formatted_tags.append({'key': tag['key'],
                                   'value': tag['value']})
        if formatted_tags:
            # NOTE(ft): AWS returns tagSet element for all objects (there are
            # errors in AWS docs)
            formatted_item['tagSet'] = formatted_tags

    def describe(self, context, ids=None, names=None, filter=None,
                 max_results=None, next_token=None):
        if filter:
            for f in filter:
                if f['name'].startswith('tag:'):
                    tag_key = f['name'].split(':')[1]
                    tag_values = f['value']
                    f['name'] = 'tag'
                    f['value'] = [{'key': tag_key,
                                   'value': tag_values}]
        return super(TaggableItemsDescriber, self).describe(
            context, ids=ids, names=names, filter=filter,
            max_results=max_results, next_token=next_token)

    def is_filtering_value_found(self, filter_value, value):
        if isinstance(filter_value, dict):
            for tag_pair in value:
                if (not isinstance(tag_pair, dict) or
                        filter_value.get('key') != tag_pair.get('key')):
                    continue
                for filter_dict_value in filter_value.get('value'):
                    if super(TaggableItemsDescriber,
                             self).is_filtering_value_found(
                                filter_dict_value,
                                tag_pair.get('value')):
                        return True
            return False
        return super(TaggableItemsDescriber,
                     self).is_filtering_value_found(filter_value, value)


class NonOpenstackItemsDescriber(UniversalDescriber):
    """Describer class for non-Openstack items Describe implementations."""

    def describe(self, context, ids=None, names=None, filter=None,
                 max_results=None, next_token=None):
        if max_results and max_results < 5:
            msg = (_('Value ( %s ) for parameter maxResults is invalid. '
                     'Expecting a value greater than 5.') % max_results)
            raise exception.InvalidParameterValue(msg)

        self.context = context
        self.ids = ids
        self.items = self.get_db_items()
        formatted_items = []

        for item in self.items:
            formatted_item = self.format(item)
            self.post_format(formatted_item, item)
            if (formatted_item and
                    not self.filtered_out(formatted_item, filter)):
                formatted_items.append(formatted_item)

        return self.get_paged(formatted_items, max_results, next_token)



# Copyright 2014
# The Cloudscaling Group, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import netaddr
from oslo_log import log as logging
import six

from ec2api import exception
from ec2api.i18n import _


LOG = logging.getLogger(__name__)


def validate_str(val, parameter_name, max_length=None):
    if (isinstance(val, six.string_types) and
            (max_length is None or max_length and len(val) <= max_length)):
        return True
    raise exception.ValidationError(
        reason=_("%s should not be greater "
                 "than 255 characters.") % parameter_name)


def validate_bool(val, parameter_name):
    if isinstance(val, bool):
        return True
    raise exception.ValidationError(
        reason=_("Expected a boolean value for parameter %s") % parameter_name)


def validate_int(val, parameter_name):
    if isinstance(val, int):
        return True
    raise exception.ValidationError(
        reason=(_("Expected an integer value for parameter %s") %
                parameter_name))


def validate_list(items, parameter_name):
    if not isinstance(items, list):
        raise exception.InvalidParameterValue(
            value=items,
            parameter=parameter_name,
            reason='Expected a list here')


def _is_valid_cidr(address):
    """Check if address is valid

    The provided address can be a IPv6 or a IPv4
    CIDR address.
    """
    try:
        # Validate the correct CIDR Address
        netaddr.IPNetwork(address)
    except netaddr.core.AddrFormatError:
        return False
    except UnboundLocalError:
        # NOTE(MotoKen): work around bug in netaddr 0.7.5 (see detail in
        # https://github.com/drkjam/netaddr/issues/2)
        return False

    # Prior validation partially verify /xx part
    # Verify it here
    ip_segment = address.split('/')

    if (len(ip_segment) <= 1 or
            ip_segment[1] == ''):
        return False

    return True


def validate_cidr_with_ipv6(cidr, parameter_name, **kwargs):
    invalid_format_exception = exception.InvalidParameterValue(
        value=cidr,
        parameter=parameter_name,
        reason='This is not a valid CIDR block.')
    if not _is_valid_cidr(cidr):
        raise invalid_format_exception
    return True


_cidr_re = re.compile(r"^([0-9]{1,3}\.){3}[0-9]{1,3}/[0-9]{1,2}$")


def validate_cidr(cidr, parameter_name):
    invalid_format_exception = exception.InvalidParameterValue(
        value=cidr,
        parameter=parameter_name,
        reason='This is not a valid CIDR block.')
    if not _cidr_re.match(cidr):
        raise invalid_format_exception
    address, size = cidr.split("/")
    octets = address.split(".")
    if any(int(octet) > 255 for octet in octets):
        raise invalid_format_exception
    size = int(size)
    if size > 32:
        raise invalid_format_exception
    return True


def _validate_cidr_block(cidr):
    validate_cidr(cidr, 'cidrBlock')
    size = int(cidr.split("/")[-1])
    return size >= 16 and size <= 28


def validate_vpc_cidr(cidr):
    if not _validate_cidr_block(cidr):
        raise exception.InvalidVpcRange(cidr_block=cidr)


def validate_subnet_cidr(cidr):
    if not _validate_cidr_block(cidr):
        raise exception.InvalidSubnetRange(cidr_block=cidr)


# NOTE(Alex) Unfortunately Amazon returns various kinds of error for invalid
# IDs (...ID.Malformed, ...Id.Malformed, ...ID.NotFound, InvalidParameterValue)
# So we decided here to commonize invalid IDs to InvalidParameterValue error.

def validate_ec2_id(val, parameter_name, prefices):
    try:
        prefix, value = val.rsplit('-', 1)
        int(value, 16)
        if not prefices or prefix in prefices:
            return True
    except Exception:
        pass

    if not prefices:
        reason = _('Invalid EC2 id was specified.')
    else:
        reason = _('Expected: %(prefix)s-...') % {'prefix': prefices[0]}
    raise exception.InvalidParameterValue(
        value=val, parameter=parameter_name, reason=reason)


def validate_ec2_association_id(id, parameter_name, action):
    if action == 'DisassociateAddress':
        return validate_ec2_id(['eipassoc'])(id, parameter_name)
    else:
        return validate_ec2_id(['rtbassoc'])(id, parameter_name)


def validate_ipv4(address, parameter_name):
    """Verify that address represents a valid IPv4 address."""
    try:
        if netaddr.valid_ipv4(address):
            return True
    except Exception:
        pass
    raise exception.InvalidParameterValue(
        value=address, parameter=parameter_name,
        reason=_('Not a valid IP address'))


def validate_enum(value, allowed_values, parameter_name, allow_empty=False):
    if value is None and allow_empty or value in allowed_values:
        return True
    raise exception.InvalidParameterValue(
        value=value, parameter=parameter_name,
        reason=_('Invalid parameter value specified'))


def validate_filter(filters):
    for filter in filters:
        if (not filter.get('name') or not filter.get('value') or
                not isinstance(filter['value'], list)):
            raise exception.InvalidFilter()
    return True


def validate_key_value_dict_list(dict_list, parameter_name):
    for dict in dict_list:
        if not dict.get('key') or dict.get('value') is None:
            raise exception.InvalidParameterValue(
                value=dict, parameter=parameter_name,
                reason=_('Expected list of key value dictionaries'))
    return True


def validate_security_group_str(value, parameter_name, vpc_id=None):
    # NOTE(Alex) Amazon accepts any ASCII for EC2 classic;
    # for EC2-VPC: a-z, A-Z, 0-9, spaces, and ._-:/()#,@[]+=&;{}!$*
    if vpc_id:
        allowed = r'^[a-zA-Z0-9\._\-:/\(\)#,@\[\]\+=&;\{\}!\$\*\ ]+$'
    else:
        allowed = r'^[\x20-\x7E]+$'
    msg = ''
    try:
        val = value.strip()
    except AttributeError:
        msg = (_("Security group %s is not a string or unicode") %
               parameter_name)
    if not val:
        msg = _("Security group %s cannot be empty.") % parameter_name
    elif not re.match(allowed, val):
        msg = (_("Specified value for parameter Group%(property)s is "
                 "invalid. Content limited to '%(allowed)s'.") %
               {'allowed': 'allowed',
                'property': parameter_name})
    elif len(val) > 255:
        msg = _("Security group %s should not be greater "
                "than 255 characters.") % parameter_name
    if msg:
        raise exception.ValidationError(reason=msg)
    return True


def validate_vpn_connection_type(value):
    if value != 'ipsec.1':
        raise exception.InvalidParameterValue(
            value=type, parameter='type',
            reason=_('Invalid VPN connection type.'))
    return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import ec2api.api
import ec2api.api.auth
import ec2api.api.availability_zone
import ec2api.api.common
import ec2api.api.dhcp_options
import ec2api.api.ec2utils
import ec2api.api.image
import ec2api.api.instance


def list_opts():
    return [
        ('DEFAULT',
         itertools.chain(
             ec2api.api.ec2_opts,
             ec2api.api.auth.auth_opts,
             ec2api.api.availability_zone.availability_zone_opts,
             ec2api.api.common.ec2_opts,
             ec2api.api.dhcp_options.ec2_opts,
             ec2api.api.ec2utils.ec2_opts,
             ec2api.api.image.s3_opts,
             ec2api.api.image.rpcapi_opts,
             ec2api.api.instance.ec2_opts,
         )),
    ]
